from tabml.model_wrappers import (
    CatBoostRegressorModelWrapper,
    LgbmRegressorModelWrapper,
    XGBoostRegressorModelWrapper,
)
from tabml.pipelines import BasePipeline


class CustomModelWrapperLog10:
    def predict(self, data):
        # In prediction, the model should return the outputs that are in the same
        # space with the true label.
        return 10 ** self.model.predict(data)


class CustomLgbmRegressorModelWrapperLog10(
    LgbmRegressorModelWrapper, CustomModelWrapperLog10
):
    pass


class CustomXGBoostRegressorModelWrapperLog10(
    XGBoostRegressorModelWrapper, CustomModelWrapperLog10
):
    pass


class CustomCatBoostRegressorModelWrapperLog10(
    CatBoostRegressorModelWrapper, CustomModelWrapperLog10
):
    pass


def train_lgbm():
    path_to_config = "configs/lgbm_config.yaml"
    pipeline = BasePipeline(
        path_to_config, custom_model_wrapper=CustomLgbmRegressorModelWrapperLog10
    )
    pipeline.run()


def train_xgboost():
    path_to_config = "configs/xgboost_config.yaml"
    pipeline = BasePipeline(
        path_to_config, custom_model_wrapper=CustomXGBoostRegressorModelWrapperLog10
    )
    pipeline.run()


def train_catboost():
    path_to_config = "configs/catboost_config.yaml"
    pipeline = BasePipeline(
        path_to_config, custom_model_wrapper=CustomCatBoostRegressorModelWrapperLog10
    )
    pipeline.run()


if __name__ == "__main__":
    train_lgbm()



from abc import ABC
from pathlib import Path

import mlflow

from tabml import experiment_manager, model_wrappers
from tabml.config_helpers import parse_pipeline_config
from tabml.data_loaders import BaseDataLoader
from tabml.model_analysis import ModelAnalysis
from tabml.utils import factory
from tabml.utils.logger import logger


class BasePipeline(ABC):
    """Base class for pipeline.

    Attributions:
        config:
            the pb object of the prototext config file
            (passed via path_to_config when initialized)
        data_loader:
            a data loader object, a subclass of base.data_loaders.BaseDataLoader.
        exp_manager:
            a base.experiment_manager.ExperimentManager object.
        custom_model_wrapper:
            model_wrapper defined by users.
    """

    def __init__(
        self, path_to_config: str, custom_model_wrapper=None, custom_run_dir=""
    ):
        logger.info("=" * 80)
        logger.info(f"Running pipeline with config {path_to_config}")
        logger.info("=" * 80)
        self.exp_manager = experiment_manager.ExperimentManger(
            path_to_config, custom_run_dir=Path(custom_run_dir)
        )
        self.config = parse_pipeline_config(path_to_config)
        self.data_loader = self._get_data_loader()
        assert self.data_loader.label_col is not None, "label_col must be specified"
        self._get_model_wrapper(custom_model_wrapper)

        logger.add(self.exp_manager.get_log_path())

    def run(self):
        self.exp_manager.create_new_run_dir()
        # start mlflow auto log
        model_type = self.model_wrapper.mlflow_model_type
        model_wrappers.MLFLOW_AUTOLOG[model_type]
        with mlflow.start_run():
            mlflow.log_param("model_type", model_type)
            mlflow.log_params(
                {
                    "model_params": self.config.model_wrapper.model_params,
                    "fit_params": self.config.model_wrapper.fit_params,
                }
            )
            self.train()
            self.analyze_model()

    def train(self):
        model_dir = self.exp_manager.run_dir
        logger.info("Start training the model.")
        self.model_wrapper.fit(self.data_loader, model_dir)

    def _get_data_loader(self) -> BaseDataLoader:
        return factory.create(self.config.data_loader.cls_name)(self.config.data_loader)

    def _get_model_wrapper(self, custom_model_wrapper):
        if custom_model_wrapper:
            self.model_wrapper = custom_model_wrapper(self.config.model_wrapper)
        else:
            self.model_wrapper = factory.create(self.config.model_wrapper.cls_name)(
                self.config.model_wrapper
            )

    def analyze_model(self) -> None:
        """Analyze the model on the validation dataset.

        The trained model is evaluated based on metrics for predictions slicing by
        each categorical feature specified by features_to_analyze.
        """
        logger.info("Model Analysis")
        assert len(self.config.model_analysis.metrics) > 0, (
            "At least one metrics in model_analysis must be specified. "
            "Add the metrics in model_analysis in the pipeline config"
        )
        assert len(self.config.model_analysis.by_features) > 0, (
            "At least one by_features in model_analysis must be specified. "
            "Add the by_features in model_analysis in the pipeline config"
        )

        ModelAnalysis(
            data_loader=self.data_loader,
            model_wrapper=self.model_wrapper,
            params=self.config.model_analysis,
            output_dir=self.exp_manager.get_model_analysis_dir(),
        ).analyze()



import importlib


def create(cls):
    """Creates a class based on its name."""
    module_name, class_name = cls.rsplit(".", 1)
    somemodule = importlib.import_module(module_name)
    cls_instance = getattr(somemodule, class_name)

    return cls_instance



import pandas as pd
import pytest
from qcore.asserts import assert_eq

from tabml import data_processing
from tabml.utils.utils import write_str_to_file


class TestQuantileClipper:
    def test_no_nan(self):
        # 5
        nums = 5 * [-1.0] + 90 * [1] + 5 * [99]
        df = pd.DataFrame(data={"col": nums})
        expected_data = 5 * [0.0] + 90 * [1] + 5 * [50]
        expected = pd.Series(data=expected_data)
        # got = QuantileClipper(lower_percentile=.05, upper_percentile=.95).fit(
        #     X=df["col"]).transform(df["col"])
        got = data_processing.QuantileClipper(
            lower_percentile=0.05, upper_percentile=0.95, interpolation="midpoint"
        ).fit_transform(df["col"])
        pd.testing.assert_series_equal(expected, got, check_names=False)


class TestFindBoxPlotBoundaries:
    def test_1(self):
        df = pd.DataFrame(data={"a": ["a", "b", "c", "d"], "val": [1, 2, 3, 4]})
        got = data_processing.find_boxplot_boundaries(df["val"])
        # Q1 = 1.75, Q3 = 3.25, IQR = 1.5
        expected = -0.5, 5.5
        assert_eq(expected, got)

    def test_2(self):
        df = pd.DataFrame(
            data={"val": [-10, 1, 2, 3, 4, 10], "val2": [-10, 1, 2, 3, 4, 10]}
        )
        got = pd.DataFrame(
            data=data_processing.BoxplotOutlierClipper().fit_transform(df),
            columns=["val", "val2"],
        )
        expected = pd.DataFrame(
            data={"val": [-2.5, 1, 2, 3, 4, 7.5], "val2": [-2.5, 1, 2, 3, 4, 7.5]}
        )
        pd.testing.assert_frame_equal(expected, got)


class TestCrossColumns:
    def test_1(self):
        df = pd.DataFrame(data={"a": ["x", "y", "z"], "b": [1, 2, 3]})
        expected = pd.Series(data=["x_X_1", "y_X_2", "z_X_3"])
        got = data_processing.cross_columns(df, cols=["a", "b"])
        pd.testing.assert_series_equal(expected, got, check_names=False)


class TestCategoryEncoder:
    @pytest.fixture(autouse=True)
    def setup_class(self, tmp_path):
        self.series = pd.Series(["a", "b", "c", "d"])

    def test_1(self):
        vocab_map = {"a": 0, "b": 1, "c": 2, "d": 3}
        got = data_processing.CategoryEncoder(vocab_map).get_encoded(self.series)
        expected = pd.Series([0, 1, 2, 3])
        pd.testing.assert_series_equal(expected, got)

    def test_unknown(self):
        vocab_map = {"a": 0, "b": 1, "c": 5}
        got = data_processing.CategoryEncoder(vocab_map).get_encoded(self.series)
        expected = pd.Series([0, 1, 5, 6])
        pd.testing.assert_series_equal(expected, got)

    def test_from_list(self):
        vocab = ["a", "c", "b"]
        got = data_processing.CategoryEncoder.from_list(vocab).get_encoded(self.series)
        expected = pd.Series([0, 2, 1, 3])
        pd.testing.assert_series_equal(expected, got)

    def test_from_lines_in_txt(self, tmp_path):
        file_content = """a
            d
            c
        """
        txt_path = tmp_path / "foo.txt"
        write_str_to_file(file_content, txt_path)
        got = data_processing.CategoryEncoder.from_lines_in_txt(txt_path).get_encoded(
            self.series
        )
        expected = pd.Series([0, 3, 2, 1])
        pd.testing.assert_series_equal(expected, got)

    def test_from_mapping_in_csv(self, tmp_path):
        file_content = """
            a,0
            d, 2
            c,5
        """
        txt_path = tmp_path / "bar.txt"
        write_str_to_file(file_content, txt_path)
        got = data_processing.CategoryEncoder.from_mapping_in_csv(txt_path).get_encoded(
            self.series
        )
        expected = pd.Series([0, 6, 5, 2])
        pd.testing.assert_series_equal(expected, got)



from tabml.utils.utils import change_working_dir_pytest

from . import feature_manager, pipelines


@change_working_dir_pytest
def test_feature_manager():
    feature_manager.run()


@change_working_dir_pytest
def test_full_pipeline_lgbm():
    pipelines.train_lgbm()


@change_working_dir_pytest
def test_full_pipeline_xgboost():
    pipelines.train_xgboost()


@change_working_dir_pytest
def test_full_pipeline_catboost():
    pipelines.train_catboost()



"""Utilities that support making inference."""

from typing import Any, Dict, List, Union

from tabml.config_helpers import parse_pipeline_config
from tabml.experiment_manager import ExperimentManger
from tabml.utils import factory


class ModelInference:
    def __init__(
        self,
        feature_manager_cls,
        feature_config_path: str,
        model_path: str,
        pipeline_config_path: Union[str, None] = None,
        custom_model_wrapper=None,
        transformer_path=None,
    ):
        self.fm = feature_manager_cls(
            feature_config_path, transformer_path=transformer_path
        )
        self.fm.load_transformers()
        if pipeline_config_path is None:
            pipeline_config_path = ExperimentManger.get_config_path_from_model_path(
                model_path
            )
        self.config = parse_pipeline_config(pipeline_config_path)
        if custom_model_wrapper:
            self.model_wrapper = custom_model_wrapper(self.config.model_wrapper)
        else:
            self.model_wrapper = factory.create(self.config.model_wrapper.cls_name)(
                self.config.model_wrapper
            )
        self.model_wrapper.load_model(model_path)
        self.features_to_model = list(self.config.data_loader.features_to_model)

    def predict(self, raw_data: List[Dict[str, Any]]) -> List[Any]:
        """Makes prediction for new samples.

        The sample should be given in a form of a list of dictionaries whose keys are
        the raw features. The feature manager transform the raw_data into a processed
        dataframe that is fed into the model to make predictions.
        """
        features = self.fm.transform_new_samples(raw_data, self.features_to_model)
        return self.model_wrapper.predict(features[self.features_to_model])



import pytest
from qcore.asserts import AssertRaises, assert_eq

from tabml import feature_config_helper
from tabml.config_helpers import parse_feature_config
from tabml.utils.utils import write_str_to_file


class TestFeatureConfigHelper:
    @pytest.fixture(autouse=True)
    def setup_class(cls, tmp_path):
        feature_config_str = """
            raw_data_dir: "dummy"
            dataset_name: "dummy"
            base_features:
              - name: "a"
                dtype: STRING
            transforming_features:
              - name: "b"
                index: 1
                dtype: STRING
                dependencies:
                  - "a"
              - name: "c"
                index: 2
                dtype: STRING
                dependencies:
                  - "a"
                  - "b"
              - name: "d"
                index: 3
                dtype: STRING
                dependencies:
                  - "a"
              - name: "e"
                index: 4
                dtype: STRING
                dependencies:
                  - "c"
        """
        config_path = tmp_path / "feature_config_str.yaml"
        write_str_to_file(feature_config_str, config_path)
        cls.fm_helper = feature_config_helper.FeatureConfigHelper(config_path)

    def test_raise_value_error_with_invalid_indexes(self, tmp_path):
        invalid_index_STRING = """
            # invalid config with indexes are not continuous
            raw_data_dir: "dummy"
            dataset_name: "dummy"
            base_features:
              - name: "TIME"
                dtype: "DATETIME"
            transforming_features:
              - name: "weekday"
                index: 1
                dtype: INT32
                dependencies:
                  - "TIME"
              - name: "hour"
                index: 1
                dtype: INT32
                dependencies:
                  - "TIME"
        """
        config_path = tmp_path / "tmp.yaml"
        write_str_to_file(invalid_index_STRING, config_path)
        with AssertRaises(ValueError) as assert_raises:
            feature_config_helper.FeatureConfigHelper(config_path)

        error_message = assert_raises.expected_exception_found
        assert_eq(
            True,
            error_message.args[0].startswith(
                "Feature indexes must be a list of increasing positive integers. "
                "Got indexes = [1, 1]"
            ),
        )

    def test_raise_assertion_error_with_duplicate_features(self, tmp_path):
        config_STRING = """
            raw_data_dir: "dummy"
            dataset_name: "dummy"
            base_features:
              - name: "TIME"
                dtype: DATETIME
            transforming_features:
              - name: "weekday"
                index: 1
                dtype: STRING
                dependencies:
                - "TIME"
              - name: "weekday"
                index: 2
                dtype: STRING
        """
        config_path = tmp_path / "tmp.pb"
        write_str_to_file(config_STRING, config_path)
        with AssertRaises(AssertionError) as assert_raises:
            feature_config_helper.FeatureConfigHelper(config_path)

        error_message = assert_raises.expected_exception_found
        assert_eq(
            True,
            error_message.args[0].startswith(
                "There are duplicate objects in the list: "
            ),
        )

    def test_raise_value_error_with_invalid_dependencies(self, tmp_path):
        invalid_dependency_STRING = """
            raw_data_dir: "dummy"
            dataset_name: "dummy"
            base_features:
              - name: "TIME"
                dtype: DATETIME
            transforming_features:
              - name: "weekday"
                index: 1
                dtype: STRING
                dependencies:
                  - "date"
        """
        config_path = tmp_path / "tmp.pb"
        write_str_to_file(invalid_dependency_STRING, config_path)
        with AssertRaises(AssertionError) as assert_raises:
            feature_config_helper.FeatureConfigHelper(config_path)

        error_message = assert_raises.expected_exception_found
        assert_eq(
            True,
            error_message.args[0].startswith(
                "Feature weekday depends on feature date that is undefined."
            ),
        )

    def test_find_dependents(self, tmp_path):
        got_1 = self.fm_helper.find_dependents("a")
        expected_1 = ["b", "c", "d", "e"]
        assert_eq(expected_1, got_1)

        got_2 = self.fm_helper.find_dependents("b")
        expected_2 = ["c", "e"]
        assert_eq(expected_2, got_2)

        got_3 = self.fm_helper.find_dependents("d")
        expected_3 = []
        assert_eq(expected_3, got_3)

    def test_append_dependents(self, tmp_path):
        got = self.fm_helper.append_dependents(["d", "b"])
        expected = ["b", "c", "d", "e"]
        assert_eq(expected, got)

    def test_extract_config_1(self, tmp_path):
        subset_features = ["e"]
        expected_STRING = """
            raw_data_dir: "dummy"
            dataset_name: "dummy"
            base_features:
              - name: "a"
                dtype: STRING
            transforming_features:
              - name: "b"
                index: 1
                dtype: STRING
                dependencies:
                  - "a"
              - name: "c"
                index: 2
                dtype: STRING
                dependencies:
                  - "a"
                  - "b"
              - name: "e"
                index: 4
                dtype: STRING
                dependencies:
                  - "c"
        """
        new_config_path = str(tmp_path / "new_tmp.yaml")
        write_str_to_file(expected_STRING, new_config_path)
        new_config = self.fm_helper.extract_config(selected_features=subset_features)
        assert_eq(parse_feature_config(new_config_path), new_config)

    def test_extract_config_2(self, tmp_path):
        subset_features = ["d"]
        expected_STRING = """
            raw_data_dir: "dummy"
            dataset_name: "dummy"
            base_features:
              - name: "a"
                dtype: STRING
            transforming_features:
              - name: "d"
                index: 3
                dtype: STRING
                dependencies:
                  - "a"
        """
        new_config_path = str(tmp_path / "new_tmp.yaml")
        write_str_to_file(expected_STRING, new_config_path)
        new_config = self.fm_helper.extract_config(selected_features=subset_features)
        assert_eq(parse_feature_config(new_config_path), new_config)

    def test_raise_value_error_with_invalid_feature_to_extract(self, tmp_path):
        subset_features = ["a", "y", "z"]
        with AssertRaises(ValueError) as assert_raises:
            self.fm_helper.extract_config(selected_features=subset_features)

        error_message = assert_raises.expected_exception_found
        assert_eq(
            error_message.args[0], "Features ['y', 'z'] are not in the original config."
        )



from pathlib import Path
from typing import Any, Dict, Iterable, List, Union

import mlflow
import pandas as pd

from tabml.data_loaders import BaseDataLoader
from tabml.metrics import BaseMetric, get_instantiated_metric_dict
from tabml.model_wrappers import BaseModelWrapper
from tabml.schemas import pipeline_config
from tabml.utils import utils
from tabml.utils.logger import logger


class ModelAnalysis:
    """A class performing model analysis on validation dataset on different dimensions.

    For each dimension in features_to_analyze (must be a categorical feature),
    group samples by each possible value then compute the metric scores for the groups.
    Results are then saved in to a csv file with header:
    feature_name, sample_count, metric_0, metric_1, ....
    Rows in each dataframe are sorted by metric_0 column from worst to best (since we
    often want to know how the model performance on the hard groups). Results are saved
    into an output directory.
    Repeat for all dimensions.

    Attributes:
        data_loader:
            A data loader object.
        model_wrapper:
            A model wrapper object.
        features_to_analyze:
            A list of feature names to do the analysis.
        label_to_analyze:
            A string of label name for the analysis. This label could be different from
            training label.
        metrics:
            A list of metric names to be computed.
        output_dir:
            A string of output directory.
        pred_col:
            A string of prediction column (default to "prediction").
        pred_prob_col:
            A string of prediction probability column.
        need_predict:
            A bool value to tell if method self.model_wrapper.predict is needed.
        need_predict_proba:
            A bool value to tell if method self.model_wrapper.predict_proba is needed.
            Some metrics are computed based on probabilities.
        training_size: int or None
            Training sample size for the analysis.
    """

    def __init__(
        self,
        data_loader: BaseDataLoader,
        model_wrapper: BaseModelWrapper,
        params=pipeline_config.ModelAnalysis(),
        output_dir: str = "",
        pred_col: str = "prediction",
        pred_proba_col: str = "prediction_probability",
    ):

        self.data_loader = data_loader
        self.model_wrapper = model_wrapper
        self.features_to_analyze = params.by_features
        self.metrics = _get_metrics(params.metrics)
        self.output_dir = output_dir
        self.label_to_analyze = params.by_label or self.data_loader.label_col
        self.pred_col = pred_col
        self.pred_proba_col = pred_proba_col
        need_pred_proba_list = [metric.need_pred_proba for metric in self.metrics]
        self.need_predict_proba = any(need_pred_proba_list)
        self.need_predict = any(
            not need_pred_proba for need_pred_proba in need_pred_proba_list
        )
        # A bool value to determine when to print the overall scores. This is set to
        # True at initialization, then to False right after printing the first overall
        # scores.
        self._show_overall_flag = True
        self.training_size = params.training_size

    def analyze(self):
        self._show_feature_importance()
        self._analyze_metrics_one_dataset("train")
        self._analyze_metrics_one_dataset("val")

    def _show_feature_importance(self):
        train_feature = self._get_dataset("train")[self.data_loader.features]
        feature_importance_dict = self.model_wrapper.get_feature_importance(
            train_feature
        )
        utils.show_feature_importance(feature_importance_dict)

    def _analyze_metrics_one_dataset(self, dataset_name: str):
        self._show_overall_flag = True
        dataset = self._get_dataset(dataset_name)
        preds = self._get_predictions(dataset)
        df_with_pred = self._get_df_with_pred(dataset, preds)
        for feature_name in self.features_to_analyze:
            self._analyze_on_feature(df_with_pred, feature_name, dataset_name)

    def _get_dataset(self, dataset_name: str):
        all_features = (
            list(self.features_to_analyze)
            + self.data_loader.features
            + [self.label_to_analyze]
        )
        if dataset_name == "train":
            df = self.data_loader.feature_manager.extract_dataframe(
                features_to_select=all_features, filters=self.data_loader.train_filters
            )
            if self.training_size:
                return df.sample(n=min(int(self.training_size), len(df)))
            return df

        if dataset_name == "val":
            return self.data_loader.feature_manager.extract_dataframe(
                features_to_select=all_features,
                filters=self.data_loader.validation_filters,
            )
        raise ValueError(f"dataset_name ({dataset_name}) not in ('train', 'val')")

    def _get_predictions(self, dataset) -> Dict[str, Iterable]:
        res = {}
        if self.need_predict:
            res[self.pred_col] = self.model_wrapper.predict(
                dataset[self.data_loader.features]
            )
        if self.need_predict_proba:
            res[self.pred_proba_col] = self.model_wrapper.predict_proba(
                dataset[self.data_loader.features]
            )
        return res

    def _get_df_with_pred(self, df, preds) -> pd.DataFrame:
        """Appends prediction column to the dataframe."""
        assert (
            self.pred_col not in df.columns
        ), f"{self.pred_col} column already exists in dataset."
        if self.need_predict:
            df[self.pred_col] = preds[self.pred_col]
        if self.need_predict_proba:
            df[self.pred_proba_col] = preds[self.pred_proba_col]
        df.to_csv(self._get_df_pred_csv_path(), index=False)
        return df

    def _analyze_on_feature(
        self, df_with_pred: pd.DataFrame, feature_name: str, dataset_name: str
    ):
        """Analyzes the predictions based on one feature."""
        # get list of metric score dict for each value in feature_name
        list_of_group_dicts = [
            self._get_metric_dict(feature_name, feature_value, group)
            for feature_value, group in df_with_pred.groupby(feature_name)
        ]
        # add overall score
        overall_scores = self._get_metric_dict(feature_name, "OVERALL", df_with_pred)
        list_of_group_dicts.append(overall_scores)
        if self._show_overall_flag:
            self._show_and_log_overall_scores(overall_scores, dataset_name)
            self._show_overall_flag = False

        df_group = pd.DataFrame(list_of_group_dicts).sort_values(
            by=self.metrics[0].score_names[0],
            ascending=self.metrics[0].is_higher_better,
        )
        saved_path = self._get_df_feature_metric_csv_path(dataset_name, feature_name)
        df_group.to_csv(saved_path, index=False)
        logger.info(
            f"Saved model analysis slicing against {feature_name} to {saved_path}"
        )

    def _get_metric_dict(
        self, feature_name: str, feature_value: Any, df_with_pred: pd.DataFrame
    ) -> Dict[str, Union[str, float]]:
        labels = df_with_pred[self.label_to_analyze]

        res = {feature_name: feature_value, "sample_count": len(df_with_pred)}
        for metric in self.metrics:
            if metric.need_pred_proba:
                res.update(
                    metric.compute_scores(labels, df_with_pred[self.pred_proba_col])
                )
            else:
                res.update(metric.compute_scores(labels, df_with_pred[self.pred_col]))
        return res

    def _get_df_pred_csv_path(self):
        return Path(self.output_dir).joinpath("prediction.csv")

    def _get_df_feature_metric_csv_path(self, dataset_name: str, col: str):
        dirname = Path(self.output_dir) / dataset_name
        if not dirname.exists():
            dirname.mkdir()

        return dirname / f"{col}.csv"

    def _show_and_log_overall_scores(
        self, overall_scores: Dict[str, Any], dataset_name: str
    ) -> None:
        logger.info("=" * 20 + f" OVERALL SCORES on {dataset_name} dataset " + "=" * 20)
        logger.info("{:<20}: {}".format("Num samples", overall_scores["sample_count"]))
        for key, val in overall_scores.items():
            if val == "OVERALL" or key == "sample_count":
                continue
            logger.info("{:<20}: {}".format(key, val))
            mlflow.log_metrics({key: val})


def _get_metrics(metric_names: List[str]) -> List[BaseMetric]:
    metric_by_name = get_instantiated_metric_dict()
    return [metric_by_name[metric_name] for metric_name in metric_names]



"""Embedding utilities."""

import numpy as np


class NearestNeighbor:
    """Class supporting finding neareast embeddings of a query embeddings.

    Attrubutes:
        item_embeddings: a matrix of shape [N, k], such that row i is the embedding of
            item i.
        measure: a string specifying the similarity measure to be used.
            One of ("cosine", "dot", "l2")
    """

    def __init__(self, item_embeddings, measure="cosine"):
        assert measure in ("dot", "cosine", "l2")
        self.measure = measure
        self.item_embeddings = item_embeddings
        if self.measure == "cosine":
            # nomalize embeding
            self.item_embeddings = item_embeddings / np.linalg.norm(
                item_embeddings, axis=1, keepdims=True
            )
        elif self.measure == "l2":
            self.squared_item_embedding = (item_embeddings ** 2).sum(axis=1)

    def find_nearest_neighbors(self, query_embedding, k=10):
        """Returns indices of k nearest neighbors"""
        # Denote q as query_emebdding vector, V as item_embeddings matrix.
        dot_products = query_embedding.dot(self.item_embeddings.T)
        if self.measure in ("dot", "cosine"):
            scores = dot_products
        elif self.measure == "l2":
            # ignore squared_query_embedding since it's the same for all item_embeddings
            scores = -(self.squared_item_embedding - 2 * dot_products)

        return (-scores).argsort()[:k]



import copy
from typing import Dict, List, Union

from tabml.config_helpers import parse_feature_config
from tabml.schemas.feature_config import DType
from tabml.utils.utils import check_uniqueness


class _Feature:
    def __init__(
        self, index: int, dtype: DType, dependencies: Union[List, None] = None
    ):
        self.index = index
        self.dependents: List[str] = []
        self.dtype = dtype
        if dependencies is None:
            self.dependencies = []
        else:
            self.dependencies = dependencies


class FeatureConfigHelper:
    """A config helper class for feature manager.

    Attributes:
        _config:
            A protobuf object parsed from a pb text file.
        raw_data_dir:
            A string of directory to raw data.
        dataset_name:
            A string of the dataset name
        base_features:
            A list of base feature names in the config.
        transforming_features:
            A list of transforming feature names in the config.
        all_features:
            A list of all features in the config.
        feature_metadata:
            A dict of {feature_name: (index, dtype, list of its dependents)}.
            This is useful when finding all dependents of one feature.
    """

    def __init__(self, config_path: str):
        self._config = parse_feature_config(config_path)
        self.raw_data_dir = self._config.raw_data_dir
        self.dataset_name = self._config.dataset_name
        self.base_features = [feature.name for feature in self._config.base_features]
        self.transforming_features = [
            transforming_feature.name
            for transforming_feature in self._config.transforming_features
        ]
        self.all_features = self.base_features + self.transforming_features
        self._validate()
        self.feature_metadata: Dict[str, _Feature] = {}
        self._build_feature_metadata()

    def _validate(self):
        self._validate_indexes()
        self._validate_dependency()
        check_uniqueness(self.all_features)

    def _validate_indexes(self):
        """Checks if indexes are positive and monotonically increasing."""
        indexes = [
            # TODO: why we need int() here? Does yaml parse this incorrectly?
            int(transforming_feature.index)
            for transforming_feature in self._config.transforming_features
        ]
        if not (
            indexes[0] > 0
            and all(
                [
                    index_i < index_ip1
                    for (index_i, index_ip1) in zip(indexes[:-1], indexes[1:])
                ]
            )
        ):
            raise ValueError(
                "Feature indexes must be a list of increasing positive integers. "
                f"Got indexes = {indexes}"
            )

    def _validate_dependency(self):
        """Checks if all dependencies of a transforming feature exists."""
        # initialize from base_features then gradually adding transforming_feature
        features_so_far = self.base_features.copy()
        for feature in self._config.transforming_features:
            for dependency in feature.dependencies:
                assert (
                    dependency in features_so_far
                ), "Feature {} depends on feature {} that is undefined.".format(
                    feature.name, dependency
                )
            features_so_far.append(feature.name)

    def _build_feature_metadata(self):
        for feature in self._config.base_features:
            # all base features are considered to have index 0
            self.feature_metadata[feature.name] = _Feature(index=0, dtype=feature.dtype)

        for feature in self._config.transforming_features:
            self.feature_metadata[feature.name] = _Feature(
                index=feature.index,
                dtype=feature.dtype,
                dependencies=feature.dependencies,
            )
            for dependency in feature.dependencies:
                self.feature_metadata[dependency].dependents.append(feature.name)

    def sort_features(self, feature_names: List[str]) -> List[str]:
        # TODO: fix parsing index as int in yaml
        return sorted(feature_names, key=lambda x: int(self.feature_metadata[x].index))

    def find_dependencies(self, feature_name: str) -> List[str]:
        # NOTE: repeated fields in proto are NOT parsed to python lists but
        # <class 'google.protobuf.pyext._message.RepeatedScalarContainer'>.
        # pandas 1.1.4 accepts the above type as an array of column names but
        # pandas 1.3.1 does not!
        return list(self.feature_metadata[feature_name].dependencies)

    def get_dependencies_recursively(self, features: List[str]) -> List[str]:
        """Gets all dependencies of a list of features recursively.

        The input list should also be in the result.
        """
        queue = copy.copy(features)
        seen: List[str] = []
        while queue:
            feature = queue.pop(0)
            if feature in seen:
                continue
            seen.append(feature)
            queue.extend(self.find_dependencies(feature))
        return self.sort_features(seen)

    def find_dependents(self, feature_name: str) -> List[str]:
        """Finds all features that are directly/indirectly dependent on a feature.

        This is necessary when we want to update one feature. All of its dependents also
        need to be updated. The list of returning features is required to be in the
        order determined by their indexes in the proto config.

        Notes:
            * If "b" is a dependency of "a" then "a" is a dependent of "b".
            * If "a" is a dependent of "b" and "b" is a dependent of "c" then "a" is a
              dependent of "c".
        """
        # BFS to find all dependents
        dependents: List[str] = []
        queue = self.feature_metadata[feature_name].dependents
        while queue:
            feature = queue.pop(0)
            if feature in dependents:
                continue
            dependents.append(feature)
            queue.extend(self.feature_metadata[feature].dependents)
        return self.sort_features(dependents)

    def append_dependents(self, feature_names: List[str]) -> List[str]:
        """Finds all dependents of a list of features then appends them to the list.

        This will be used when multiple features need to be updated or remove. For the
        first case, all dependents also need to be updated. For the second case, it's
        required that there is no dependent in the remaining features.

        NOTE: the results should also contain the input feature_names.
        """
        all_features = feature_names
        for feature_name in feature_names:
            all_features.extend(self.find_dependents(feature_name))

        return self.sort_features(list(set(all_features)))

    def extract_config(self, selected_features: List[str]):
        """Creates a minimum valid config that contains all selected_features.

        Returns a protobuf with only a subset of transforming features and all of their
        dependencies.

        NOTE: all base features will be in the extracted config; they were created in
        the raw data loading step.

        Args:
            selected_features: a list of selected features.

        Raises:
            ValueError if selected_features contains an unknown features (not in the
            original config).
        """
        invalid_features = [
            feature for feature in selected_features if feature not in self.all_features
        ]
        if invalid_features:
            raise ValueError(
                f"Features {invalid_features} are not in the original config."
            )
        all_relevant_features = self.get_dependencies_recursively(
            features=selected_features
        )
        new_config = copy.deepcopy(self._config)
        new_config.transforming_features = [
            transforming_feature
            for transforming_feature in self._config.transforming_features
            if transforming_feature.name in all_relevant_features
        ]

        return new_config



import tempfile
from pathlib import Path

from tabml.experiment_manager import ExperimentManger
from tabml.inference import ModelInference
from tabml.utils.utils import change_working_dir_pytest

from . import feature_manager, pipelines


def _test_inference(
    config_path, test_custom_pipeline_config_path=False, transformer_path=None
):
    feature_config_path = "./configs/feature_config.yaml"
    last_model_run_dir = ExperimentManger(config_path).get_most_recent_run_dir()
    model_path = Path(last_model_run_dir) / "model_0"
    pipeline_config_path = config_path if test_custom_pipeline_config_path else None
    model_inference = ModelInference(
        feature_manager_cls=feature_manager.FeatureManager,
        feature_config_path=feature_config_path,
        model_path=model_path,
        pipeline_config_path=pipeline_config_path,
        transformer_path=transformer_path,
    )
    raw_data_samples = [
        {
            "PassengerId": 1,
            "Pclass": 1,
            "Name": "First, Mr. Last",
            "Sex": "female",
            "SibSp": 3,
            "Parch": 0,
            "Ticket": 12345,
            "Fare": 10.0,
            "Cabin": None,
            "Embarked": "C",
        },
        {
            "PassengerId": 2,
            "Pclass": 1,
            "Name": "First, Mrs. Last",
            "Sex": "male",
            "Age": 60,
            "SibSp": 0,
            "Parch": 2,
            "Ticket": 12345,
            "Fare": 100.0,
            "Cabin": None,
            "Embarked": "Q",
        },
    ]
    model_inference.predict(raw_data_samples)


@change_working_dir_pytest
def test_run():
    feature_manager.run()


@change_working_dir_pytest
def test_full_pipeline_lgbm():
    pipelines.train_lgbm()
    _test_inference("./configs/lgbm_config.yaml")


@change_working_dir_pytest
def test_full_pipeline_xgboost():
    pipelines.train_xgboost()
    _test_inference("./configs/xgboost_config.yaml")


@change_working_dir_pytest
def test_full_pipeline_catboost():
    pipelines.train_catboost()
    _test_inference(
        "./configs/catboost_config.yaml", test_custom_pipeline_config_path=True
    )


@change_working_dir_pytest
def test_full_pipeline_randomforest():
    config_path = "./configs/rf_config.yaml"
    pipelines.run(config_path)
    _test_inference(config_path, test_custom_pipeline_config_path=True)


@change_working_dir_pytest
def test_custom_transformer_path():
    with tempfile.NamedTemporaryFile() as temp:
        transformer_path = temp.name
        feature_manager.run(transformer_path)
        _test_inference("./configs/lgbm_config.yaml", transformer_path=transformer_path)



import typing

import pandas as pd

from tabml.feature_manager import BaseFeatureManager
from tabml.schemas import pipeline_config


class BaseDataLoader:
    """Base class for DataLoader.

    Support extract relevant datasets based on feature manager.
    """

    def __init__(self, params=pipeline_config.DataLoader()):
        self.label_col = params.label_col
        self.features = params.features_to_model
        self.features_and_label = self.features + [self.label_col]
        self.feature_manager = BaseFeatureManager(params.feature_config_path)
        self.feature_manager.load_dataframe()
        self.train_filters = params.train_filters
        self.validation_filters = params.validation_filters

    def _extract_data_and_label(
        self, filters: typing.List[str]
    ) -> typing.Tuple[pd.DataFrame, pd.Series]:
        """Gets data and label."""
        both = self.feature_manager.extract_dataframe(
            features_to_select=self.features_and_label, filters=filters
        )
        label = both.pop(self.label_col)
        return (both, label)

    def get_train_data_and_label(self) -> typing.Tuple[pd.DataFrame, pd.Series]:
        """Gets training data and label."""
        return self._extract_data_and_label(self.train_filters)

    def get_val_data_and_label(self):
        """Gets validation dataset."""
        return self._extract_data_and_label(self.validation_filters)



import datetime
import re
import shutil
from pathlib import Path
from typing import Union

from tabml.config_helpers import parse_pipeline_config


class ExperimentManger:
    """Class managing folder structure of an experiment.

    For each experiment, there will be one run_dir under exp_root_dir that contains
    all related information about the run, e.g. run log, config file, submission
    csv file, trained model, and a model_analysis folder.

    Attributes:
        exp_root_dir: root directory of all experiments.
        run_prefix: prefix of the run name subfolder inside exp_root_dir.
        run_dir: run dir name (run_prefix + timestamp).
    """

    log_filename = "run.log"
    config_filename = "config.pb"
    _model_analysis_dir = "model_analysis"

    def __init__(
        self,
        path_to_config: str,
        should_create_new_run_dir: bool = True,
        exp_root_dir: Path = Path("./experiments"),
        custom_run_dir: Union[None, Path] = None,
    ):
        """
        Args:
            path_to_config: path to proto pipeline config file.
            should_create_new_run_dir: create new experiment subfolder (True) or not
                (False). If not, set the experiment subfolder to the most recent run.
            run_prefix: prefix of the run name subfolder inside exp_root_dir
            run_dir: run dir name (exp_root_dir/run_prefix + timestamp)
            custom_run_dir: custom run dir that user can specify
        """
        self._path_to_config = path_to_config
        self._config = parse_pipeline_config(self._path_to_config)
        self.exp_root_dir = exp_root_dir
        self.run_prefix = self._config.config_name + "_"
        self.custom_run_dir = custom_run_dir
        if not custom_run_dir or not custom_run_dir.name:
            self.run_dir = self._get_run_dir(should_create_new_run_dir)
        else:
            self.run_dir = custom_run_dir

    def _get_run_dir(self, should_create_new_run_dir):
        if not should_create_new_run_dir:
            return self.get_most_recent_run_dir()
        return self.exp_root_dir.joinpath(self.run_prefix + _get_time_stamp())

    def create_new_run_dir(self):
        _make_dir_if_needed(self.run_dir)
        self._copy_config_file()

    def _copy_config_file(self):
        shutil.copyfile(self._path_to_config, self.get_config_path())

    def get_log_path(self):
        return self._make_path_under_run_dir(self.log_filename)

    def get_config_path(self):
        return self._make_path_under_run_dir(self.config_filename)

    def get_model_analysis_dir(self):
        res = self._make_path_under_run_dir(self._model_analysis_dir)
        _make_dir_if_needed(res)
        return res

    def _make_path_under_run_dir(self, sub_path: str) -> str:
        return self.run_dir.joinpath(sub_path)

    def get_most_recent_run_dir(self):
        """Returns the run_dir corresponding to the most recent timestamp.

        Raises:
            IOError if there is no such folder
        """
        if self.custom_run_dir:
            ValueError("get_most_recent_run_dir does not support custom run dir")
        subfolders = sorted(
            [
                sub
                for sub in self.exp_root_dir.iterdir()
                if sub.is_dir()
                and sub.name.startswith(self.run_prefix)
                and bool(re.match("[0-9]{6}_[0-9]{6}", sub.name[-13:]))
            ],
            key=lambda x: x.name[-13:],  # YYmmDD_HHMMSS
        )
        if not subfolders:
            raise IOError(
                "Could not find any run directory starting with "
                f"{self.exp_root_dir.joinpath(self.run_prefix)}"
            )
        return subfolders[-1]

    @classmethod
    def get_config_path_from_model_path(cls, model_path: str) -> str:
        run_dir = Path(model_path).parents[0]
        return str(run_dir / cls.config_filename)


def _make_dir_if_needed(dir_path: str):
    if not Path(dir_path).exists():
        Path(dir_path).mkdir(parents=True)


def _get_time_stamp() -> str:
    """Returns a time stamp string in format 'YYmmdd_HHMMSS'.

    Example: 200907_123344.
    """
    return datetime.datetime.now().strftime("%Y%m%d_%H%M%S")[2:]



import tempfile
import zipfile
from pathlib import Path
from typing import Dict

import pandas as pd
from six.moves.urllib.request import urlretrieve

from tabml.utils.logger import logger

SUPPORTED_DATASETS = ("titanic", "california_housing", "movielen-1m")
DATA_SOURCE = "https://media.githubusercontent.com/media/tiepvupsu/tabml_data/master/"


def _get_full_data_path(local_dir, tabml_data_dir, filename):
    """Returns the full path to a data file.

    Returns local path if exists, otherwise the tabml_data path.
    """
    if Path(local_dir).joinpath(filename).exists():
        return Path(local_dir).joinpath(filename)

    return f"{DATA_SOURCE}/{tabml_data_dir}/{filename}"


def load_titanic(data_dir: str) -> Dict[str, pd.DataFrame]:
    """Loads data from data_dir folder if they exist, downloads otherwise."""

    train_path = _get_full_data_path(data_dir, "titanic", "train.csv")
    test_path = _get_full_data_path(data_dir, "titanic", "test.csv")
    gender_submission_path = _get_full_data_path(
        data_dir, "titanic", "gender_submission.csv"
    )
    logger.info(
        f"Loading dataframes from {train_path}, {test_path}, and "
        f"{gender_submission_path}"
    )

    return {
        "train": pd.read_csv(train_path),
        "test": pd.read_csv(test_path),
        "gender_submission": pd.read_csv(gender_submission_path),
    }


def load_california_housing(data_dir: str) -> Dict[str, pd.DataFrame]:
    data_path = _get_full_data_path(data_dir, "california_housing", "housing.csv")
    logger.info(f"Loading dataframe from {data_path}")
    return {"housing": pd.read_csv(DATA_SOURCE + "california_housing/housing.csv")}


# TODO: manually download the dataset and load from disk
def download_movielen_1m() -> Dict[str, pd.DataFrame]:
    url = "http://files.grouplens.org/datasets/movielens/ml-1m.zip"
    # Download and extract ml-1m dataset to a temporary folder.
    tmp_dir = Path(tempfile.mkdtemp())
    tmp_file_path = tmp_dir.joinpath("tmp.zip")
    urlretrieve(url, tmp_file_path)
    with zipfile.ZipFile(tmp_file_path, "r") as tmp_zip:
        tmp_zip.extractall(tmp_dir)

    users = pd.read_csv(
        tmp_dir.joinpath("ml-1m", "users.dat"),
        delimiter="::",
        engine="python",
        names=["UserID", "Gender", "Age", "Occupation", "Zip-code"],
    )
    movies = pd.read_csv(
        tmp_dir.joinpath("ml-1m", "movies.dat"),
        delimiter="::",
        encoding="ISO-8859-1",
        engine="python",
        names=["MovieID", "Title", "Genres"],
    )

    ratings = pd.read_csv(
        tmp_dir.joinpath("ml-1m", "ratings.dat"),
        delimiter="::",
        engine="python",
        names=["UserID", "MovieID", "Rating", "Timestamp"],
        usecols=["UserID", "MovieID", "Rating", "Timestamp"],
    )

    return {"users": users, "movies": movies, "ratings": ratings}



from enum import Enum
from typing import List

import pydantic


class DType(Enum):
    INT32 = "INT32"
    BOOL = "BOOL"
    FLOAT = "FLOAT"
    INT64 = "INT64"
    STRING = "STRING"
    DOUBLE = "DOUBLE"
    # data and time types https://docs.python.org/3/library/datetime.html
    DATE = "DATE"
    TIME = "TIME"
    DATETIME = "DATETIME"


class BaseFeature(pydantic.BaseModel):
    name: str
    dtype: DType


class TransformingFeature(pydantic.BaseModel):
    # feature name, need to be unique in the FeatureConfig.
    name: str

    # index of the feature in the dataset. Indexes should be unique, positive and
    # monotonically increasing. Indexes are used to determine the feature order in the
    # whole dataset. Base features are added to the dataset first, then
    # transforming_features in the ascending order of indexes.
    index: int

    # dependencies is a list of features that are required to compute this feature.
    # This field will be used when users want to re-compute one feature. All features
    # depending on this re-computed feature are also required to be updated.
    # NOTE on terminology: If feature "a" depends on feature "b" then we call "b" is
    # one of dependencies of "a", and "a" is a dependent of "b".
    dtype: DType

    dependencies: List[str] = []


class FeatureConfig(pydantic.BaseModel):
    # directory of raw data files
    raw_data_dir: str

    # name of the dataframe to store features, also the name of the csv dataframe file.
    # The csv file path will be raw_data_dir / features / dataset_name + ".csv".
    dataset_name: str

    # base_features are features that are not dependent on any features.
    # These features are usually created right after the data cleaning step.
    base_features: List[BaseFeature]

    # transforming_features are those dependent on base features and/or other
    # transforming features. Note that the term "feature" here only apply to columns in
    # the final dataframe saved in dataset_name. Data taken from additional dataframes
    # is not considered as features.
    transforming_features: List[TransformingFeature]



import pandas as pd
import pytest

from tabml.feature_manager import BaseFeatureManager, BaseTransformingFeature
from tabml.utils.utils import write_str_to_file


class _DummyFeatureManager(BaseFeatureManager):
    """A dummy FeatureManager class to check compute_all_transforming_features."""

    def initialize_dataframe(self):
        dummy_data = {"a": [1, 2]}
        self.dataframe = pd.DataFrame(data=dummy_data, dtype="int32")
        self.dataframe["d1"] = pd.to_datetime(["2021-03-20", "2021-02-11"])

    def _get_base_transforming_class(self):
        return _DummyBaseTransformingFeature

    def load_raw_data(self):
        # required to initiate an instance of BaseFeatureManager
        raise NotImplementedError


class _DummyFeatureManager2(BaseFeatureManager):
    """A dummy FeatureManager class to check to update and clean logics."""

    def _compute_feature(self, feature_name: str, is_training: bool = True):
        print(feature_name)

    def initialize_dataframe(self):
        dummy_data = {"a": [1]}
        self.dataframe = pd.DataFrame(data=dummy_data)

    def load_raw_data(self):
        # required to initiate an instance of BaseFeatureManager
        raise NotImplementedError


class _DummyBaseTransformingFeature(BaseTransformingFeature):
    pass


class FeatureB(_DummyBaseTransformingFeature):
    name = "b"

    def transform(self, df):
        return df["a"]


class FeatureC(_DummyBaseTransformingFeature):
    name = "c"

    def transform(self, df):
        return df["b"] * 2


class FeatureD(_DummyBaseTransformingFeature):
    name = "d"

    def transform(self, df):
        return df["a"] + 1


class FeatureE(_DummyBaseTransformingFeature):
    name = "e"

    def transform(self, df):
        return df["c"] - 1


class FeatureD2(_DummyBaseTransformingFeature):
    name = "d2"

    def transform(self, df):
        return pd.DatetimeIndex(df["d1"]) + pd.DateOffset(1)


class TestBaseFeatureManager:
    @pytest.fixture(autouse=True)
    def setup_class(cls, tmp_path):
        yaml_str = """
            raw_data_dir: dummy
            dataset_name: dummy
            base_features:
              - name: a
                dtype: INT32
              - name: d1
                dtype: DATETIME
            transforming_features:
              - name: b
                index: 1
                dependencies:
                  - a
                dtype: INT32
              - name: c
                index: 2
                dependencies:
                  - b
                dtype: INT32
              - name: d
                index: 3
                dependencies:
                  - a
                dtype: INT32
              - name: e
                index: 4
                dependencies:
                  - c
                dtype: INT32
              - name: d2
                index: 5
                dependencies:
                  - d1
                dtype: DATETIME
        """
        yaml_config_path = str(tmp_path / "tmp.yaml")
        write_str_to_file(yaml_str, yaml_config_path)
        cls.fm = _DummyFeatureManager(yaml_config_path)
        cls.fm2 = _DummyFeatureManager2(yaml_config_path)
        cls.fm.initialize_dataframe()
        cls.fm2.initialize_dataframe()

    def test_compute_all_transforming_features(self, tmp_path):
        self.fm.compute_all_transforming_features()
        expected_dataframe = pd.DataFrame(
            data={
                "a": [1, 2],
                "d1": ["2021-03-20", "2021-02-11"],
                "b": [1, 2],
                "c": [2, 4],
                "d": [2, 3],
                "e": [1, 3],
                "d2": ["2021-03-21", "2021-02-12"],
            }
        ).astype(
            {
                "a": "int32",
                "b": "int32",
                "c": "int32",
                "d": "int32",
                "e": "int32",
                "d1": "datetime64[ns]",
                "d2": "datetime64[ns]",
            }
        )
        pd.testing.assert_frame_equal(expected_dataframe, self.fm.dataframe)

        # test save and load dataframe
        # make a temp dataset_path
        self.fm.dataset_path = tmp_path / "features" / "dataset.csv"
        self.fm.save_dataframe()
        # make sure dataframe is cleared
        self.fm.dataframe = None
        self.fm.load_dataframe()
        pd.testing.assert_frame_equal(expected_dataframe, self.fm.dataframe)

    def test_update_feature(self, capsys):
        self.fm2.update_feature("b")
        # check that "b", "c", "e" are computed.
        captured = capsys.readouterr()
        assert captured.out == "b\nc\ne\n"

    def test_extract_dataframe(self):
        self.fm.compute_all_transforming_features()
        features_to_select = ["a", "c", "d2"]
        filters = ["d2 > '2021-03-01'"]
        expected_dataframe = pd.DataFrame(
            data={"a": [1], "c": [2], "d2": ["2021-03-21"]}
        ).astype({"a": "int32", "c": "int32", "d2": "datetime64[ns]"})
        pd.testing.assert_frame_equal(
            expected_dataframe, self.fm.extract_dataframe(features_to_select, filters)
        )

        # test without filters
        expected_dataframe = pd.DataFrame(
            data={"a": [1, 2], "c": [2, 4], "d2": ["2021-03-21", "2021-02-12"]}
        ).astype({"a": "int32", "c": "int32", "d2": "datetime64[ns]"})
        pd.testing.assert_frame_equal(
            expected_dataframe, self.fm.extract_dataframe(features_to_select)
        )



import abc
from typing import Collection, Dict, List, Union

import numpy as np
from sklearn import metrics as sk_metrics


class BaseMetric(abc.ABC):
    """Base class for metrics.

    Usually, one metric returns one score like in the case of accuracy, rmse, mse, etc.
    However, in some cases, metrics might contain several values as precision, recall,
    f1. In these cases, method compute_scores will return a list of scores.

    Attributes:
        name:
            A string of class attribute, unique for each metric.
        is_higer_better:
            A bool value to tell if the higher the metric, the better the performance.
            The metric here is the first score in the output of compute_scores function.
            Note that, for some metrics like RMSE, lower numbers are better.
        need_pred_proba:
            A bool value to decide predict or predict_proba in model will be used.
    """

    name: str = ""
    score_names = [""]
    is_higher_better: Union[bool, None] = None
    need_pred_proba: bool = False

    def compute_scores(self, labels: Collection, preds: Collection) -> Dict[str, float]:
        if self.is_higher_better is None:
            raise ValueError("Subclasses must define is_higher_better.")
        if len(labels) != len(preds):
            raise ValueError(
                f"labels (len = {len(labels)}) and preds (len = {len(preds)}) "
                "must have the same length"
            )
        scores = self._compute_scores(labels, preds)
        if len(self.score_names) != len(scores):
            raise ValueError(
                f"self.score_names ({self.score_names}) and scores ({scores}) "
                "must have the same length."
            )
        return dict(zip(self.score_names, scores))

    def _compute_scores(self, labels: Collection, preds: Collection) -> List[float]:
        raise NotImplementedError


class MAE(BaseMetric):
    """Mean Absolute Error."""

    name = "mae"
    score_names = ["mae"]
    is_higher_better = False

    def _compute_scores(self, labels: Collection, preds: Collection) -> List[float]:
        return [sk_metrics.mean_absolute_error(labels, preds)]


class RMSE(BaseMetric):
    """Root Mean Square Error."""

    name = "rmse"
    score_names = ["rmse"]
    is_higher_better = False

    def _compute_scores(self, labels: Collection, preds: Collection) -> List[float]:
        return [sk_metrics.mean_squared_error(labels, preds) ** 0.5]


class AccuracyScore(BaseMetric):
    """Accuracy for classification."""

    name = "accuracy_score"
    score_names = ["accuracy_score"]
    is_higher_better = True

    def _compute_scores(self, labels: Collection, preds: Collection) -> List[float]:
        return [sk_metrics.accuracy_score(labels, preds)]


class RocAreaUnderTheCurve(BaseMetric):
    """Area ROC under the curve for binary classification."""

    name = "roc_auc"
    score_names = ["roc_auc"]
    is_higher_better = True
    need_pred_proba = True

    def _compute_scores(self, labels: Collection, preds: Collection) -> List[float]:
        # In some small groups of samples, labels might be all 0 or 1, which might cause
        # ValueError when computing ROC AUC.
        try:
            res = sk_metrics.roc_auc_score(labels, preds)
        except ValueError as error_message:
            if (
                "Only one class present in y_true. "
                "ROC AUC score is not defined in that case." in repr(error_message)
            ):
                res = np.NaN
            else:
                raise ValueError(error_message)
        return [res]


class F1Score(BaseMetric):
    """F1 score."""

    name = "f1"
    score_names = ["f1", "precision", "recall"]
    is_higher_better = True
    need_pred_proba = False

    def _compute_scores(self, labels: Collection, preds: Collection) -> List[float]:
        _check_binary_list(labels)
        _check_binary_list(preds)
        labels = np.array(labels)
        preds = np.array(preds)
        eps = 1e-8
        tps = np.sum(np.logical_and(preds == 1, labels == 1))
        fps = np.sum(np.logical_and(preds == 1, labels == 0))
        fns = np.sum(np.logical_and(preds == 0, labels == 1))
        eps = 1e-8
        precision = tps / np.maximum(tps + fps, eps)
        recall = tps / np.maximum(tps + fns, eps)
        f1 = 2 * precision * recall / np.maximum(precision + recall, eps)
        return [f1, precision, recall]


class MaxF1(BaseMetric):
    """Maximum F1 score accross multiple thresholds."""

    name = "max_f1"
    score_names = ["max_f1", "max_f1_threshold", "max_f1_precision", "max_f1_recall"]
    is_higher_better = True
    need_pred_proba = True

    def _compute_scores(self, labels: Collection, preds: Collection) -> List[float]:
        return compute_maxf1_stats(labels, preds)


def compute_maxf1_stats(labels: Collection, preds: Collection) -> List[float]:
    """Computes max f1 and the related stats for binary classification.

    Args:
        labels: An iterable variable of binary labels
        preds: An iterable variable of probability predictions. preds are clipped into
            range [0, 1] before computing f1 score

    Returns:
        A tuple of (max f1, threshold at max f1, precision at max f1, recall at max f1).

    Raises:
        ValueError if labels contains non-binary values.
    """
    num_thresholds = 1000
    labels = np.array(labels).reshape((-1, 1))  # shape (N, 1)
    preds = np.clip(np.array(preds), 0, 1).reshape((-1, 1))  # shape (N, 1)
    _check_binary_list(labels)
    thresholds = np.arange(start=0, stop=1, step=1.0 / num_thresholds)  # shape (T,)
    binary_preds = preds >= thresholds  # shape (N, T) with T = num_thresholds
    tps = np.sum(np.logical_and(binary_preds == 1, labels == 1), axis=0)
    fps = np.sum(np.logical_and(binary_preds == 1, labels == 0), axis=0)
    fns = np.sum(np.logical_and(binary_preds == 0, labels == 1), axis=0)
    eps = 1e-8
    precisions = tps / np.maximum(tps + fps, eps)
    recalls = tps / np.maximum(tps + fns, eps)
    f1s = 2 * precisions * recalls / np.maximum(precisions + recalls, eps)
    max_f1 = np.max(f1s)
    max_f1_index = np.argmax(f1s)
    max_f1_threshold = thresholds[max_f1_index]
    max_f1_precision = precisions[max_f1_index]
    max_f1_recall = recalls[max_f1_index]
    return [max_f1, max_f1_threshold, max_f1_precision, max_f1_recall]


class SMAPE(BaseMetric):
    name = "smape"  # symmetric-mean-percentage-error
    score_names = ["smape"]
    is_higher_better = False

    def _compute_scores(self, labels: Collection, preds: Collection) -> List[float]:
        nominator = 2 * np.abs(np.array(preds) - np.array(labels))
        denominator = np.abs(labels) + np.abs(preds)
        return [100 * np.mean(np.divide(nominator, denominator))]


def get_instantiated_metric_dict() -> Dict[str, BaseMetric]:
    res = {}
    for sub_class in BaseMetric.__subclasses__():
        metric = sub_class()
        res[metric.name] = metric
    return res


def _check_binary_list(nums: Collection) -> None:
    """Checks if a list only contains binary values.

    Raises an error if not.
    """
    unique_vals = np.unique(nums)
    for label in unique_vals:
        if label not in [0, 1]:
            raise ValueError(
                f"Input must contain only binary values, got {unique_vals}"
            )



import pickle
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import pandas as pd

from tabml.feature_config_helper import FeatureConfigHelper
from tabml.schemas.feature_config import DType
from tabml.utils.logger import logger
from tabml.utils.utils import check_uniqueness

PANDAS_DTYPE_MAPPING = {
    DType.BOOL: "bool",
    DType.INT32: "int32",
    DType.INT64: "int64",
    DType.STRING: "str",
    DType.FLOAT: "float32",
    DType.DOUBLE: "float64",
    # DATETIME will be converted to datetime parse_date https://tinyurl.com/y4waw6np
    DType.DATETIME: "datetime64[ns]",
}


class BaseFeatureManager(ABC):
    """A Base class for feature manager.

    Attributes:
        config_helper: A FeatureConfigHelper instance to manage feature dependency.
        raw_data_dir: A directory of raw data (set in feature manager config).
        dataset_name: Name of the final dataset (set in feature manager config).
        dataset_path: A string of the full path to where the dataframe is saved.
        features_in_config: List of features specified in feature manager config.
        raw_data:
            A dictionary with values being supporting dataframes for each dataset.
            Should be defined in method load_raw_data().
        dataframe: The main dataframe.
        feature_metadata:
            A dict of {feature_name: (index, dtype, list of its dependents)}.
        base_transforming_feature_class:
            A direct subclass of BaseTransformingFeature, is defined in each project.
        transforming_class_by_feature_name:
            A dictionary of
            {feature_name: a transforming feature generating that feature}.
            All the transforming features must be direct subclasses of
            self.base_transforming_feature_class.
        transformer_dict:
            A dictionary of {feature_name: its transformer}. This transformer_dict is
            formed and saved to a pickle in the "training" stage. In "serving" stage,
            it's loaded and does the transformations.
        transformer_path:
            pickle path to save transformers.
    """

    def __init__(self, config_path: str, transformer_path: Union[str, None] = None):
        self.config_helper = FeatureConfigHelper(config_path)
        self.raw_data_dir = self.config_helper.raw_data_dir
        self.dataset_name = self.config_helper.dataset_name
        self.feature_metadata = self.config_helper.feature_metadata
        self.dataset_path = self.get_dataset_path()
        self.features_in_config: List[str] = self.config_helper.all_features
        self.raw_data: Dict[str, Any] = {}
        self.dataframe: pd.DataFrame = pd.DataFrame()
        self.transforming_class_by_feature_name: Dict[str, Any] = {}
        self.transformer_dict: Dict[str, Any] = {}
        self.transformer_path = transformer_path or self.get_transformer_path()

    def get_dataset_path(self):
        return Path(self.raw_data_dir) / "features" / f"{self.dataset_name}.csv"

    def get_transformer_path(self):
        return Path(self.raw_data_dir) / "features" / "transformers.pickle"

    def _get_base_transforming_class(self):
        raise NotImplementedError

    def _get_transforming_class_by_name(self):
        base_transforming_feature_class = self._get_base_transforming_class()
        transforming_classes = base_transforming_feature_class.__subclasses__()
        check_uniqueness(
            [transforming_class.name for transforming_class in transforming_classes]
        )
        self.transforming_class_by_feature_name = {
            transforming_class.name: transforming_class
            for transforming_class in transforming_classes
        }

    def initialize_dataframe(self):
        """Inits the main dataframe with base features."""
        # TODO: check if the set of columns in dataframe after initialiation is exactly
        # the set of base features.
        raise NotImplementedError

    def load_raw_data(self):
        """Loads data from raw csv files and save them to self.raw_data."""
        raise NotImplementedError

    def load_dataframe(self):
        """Loads the dataframe from disk with appropriate types."""
        parse_dates = [
            feature
            for feature, metadata in self.feature_metadata.items()
            if metadata.dtype == DType.DATETIME
        ]
        self.dataframe = pd.read_csv(
            self.dataset_path,
            dtype={
                feature: PANDAS_DTYPE_MAPPING[metadata.dtype]
                for feature, metadata in self.feature_metadata.items()
                if metadata.dtype != DType.DATETIME
            },
            parse_dates=parse_dates,
        )

    def load_transformers(self):
        """Loads transformers from pickle."""
        with open(self.transformer_path, "rb") as pickle_file:
            self.transformer_dict = pickle.load(pickle_file)

    def save_dataframe(self):
        """Saves the dataframe to disk."""
        if not self.dataset_path.parent.exists():
            self.dataset_path.parent.mkdir(parents=True)
        logger.info(f"Saving dataframe to {self.dataset_path}")
        self.dataframe.to_csv(self.dataset_path, index=False)

    def save_transformers(self):
        """Saves the transformers to disk."""
        if not self.dataset_path.parent.exists():
            self.dataset_path.parent.mkdir(parents=True)
        logger.info(f"Saving transformers to {self.transformer_path}")
        with open(self.transformer_path, "wb") as pickle_file:
            pickle.dump(self.transformer_dict, pickle_file)

    def compute_feature(self, feature_name: str, is_training: bool = True) -> None:
        """Computes one feature column.

        This method should be used only when a new feature is added to the dataframe.
        If the feature exists, method update_feature should be called to make sure that
        all dependents of feature_name are also udpated.
        """
        if self.dataframe is None:
            raise NotImplementedError(
                "self.dataframe must be initialized in self.initialize_dataframe()"
            )
        assert feature_name not in self.dataframe.columns, (
            f"Feature {feature_name} already exists in the dataframe. Do you want to "
            "update it (using update_feature() method) instead?"
        )
        if not self.transforming_class_by_feature_name:
            self._get_transforming_class_by_name()
        self._compute_feature(feature_name, is_training=is_training)

    def _compute_feature(self, feature_name: str, is_training: bool = True) -> None:
        if self.dataframe is None:
            raise NotImplementedError(
                "self.dataframe must be initialized in self.initialize_dataframe()"
            )
        transforming_class = self.transforming_class_by_feature_name[feature_name]
        transformer = self.transformer_dict.get(feature_name)
        transformer_object = transforming_class(
            dependencies=self.config_helper.find_dependencies(feature_name),
            raw_data=self.raw_data,
        )
        series = transformer_object._transform(self.dataframe, transformer)
        if is_training:
            self.transformer_dict[feature_name] = transformer_object.transformer
        dtype = self.feature_metadata[feature_name].dtype
        if dtype == DType.DATETIME:
            self.dataframe.loc[:, feature_name] = pd.to_datetime(series)
        else:
            self.dataframe.loc[:, feature_name] = pd.Series(
                series, dtype=PANDAS_DTYPE_MAPPING[dtype]
            )

    def compute_all_transforming_features(
        self, transforming_features: Union[List[str], None] = None, is_training=True
    ):
        """Computes all transforming feature.

        Should be done occasionally. After the first time this method is called, it's
        expected that features are updated one by one or in a set of few features.
        """
        if transforming_features is None:
            transforming_features = self.config_helper.transforming_features
        for feature_name in transforming_features:
            self.compute_feature(feature_name, is_training=is_training)

    def update_feature(self, feature_name: str):
        """Updates one feature in the dataframe.

        If this is an existing feature, all of its dependents should be computed.
        """
        features_to_compute = [feature_name]
        features_to_compute.extend(self.config_helper.find_dependents(feature_name))
        for _feature_name in features_to_compute:
            self._compute_feature(_feature_name)

    def run_all(self):
        self.load_raw_data()
        self.initialize_dataframe()
        self.compute_all_transforming_features()
        self.save_dataframe()
        self.save_transformers()

    def transform_new_samples(self, raw_data_samples, transforming_features):
        self.set_raw_data(raw_data_samples)
        self.initialize_dataframe()
        if transforming_features is None:
            # If transforming features are not specified, get all transforming features.
            transforming_features_and_dependencies = (
                self.config_helper.transforming_features
            )
        else:
            # Find all dependencies for the transforming features. The base features
            # should be excluded since they are pre-computed in
            # self.initialize_dataframe()
            transforming_features_and_dependencies = [
                feature
                for feature in self.config_helper.get_dependencies_recursively(
                    list(transforming_features)
                )
                if feature not in self.config_helper.base_features
            ]
        self.compute_all_transforming_features(
            transforming_features_and_dependencies, is_training=False
        )
        return self.dataframe

    def set_raw_data(self, raw_data_samples: Any):
        NotImplementedError

    def extract_dataframe(
        self, features_to_select: List[str], filters: Optional[List[str]] = None
    ):
        """Extracts a subset of columns and a subset of rows in the dataframe.

        The whole idea of Feature Manager is to create a big dataframe with as many
        columns and rows as we want, which helps store the whole information in one
        place. All training/validation/test datasets are stored in this big table and
        are then extracted later per use case.

        To extract columns, we can simply use a list of column names. To extract rows,
        we can use a set of boolen filters. Training, validation and test data should
        be extracted using these filters.

        Args:
            features_to_select: A list of column names to be selected.
            filters: A list of boolen filters used to extract rows.
        """
        assert (
            self.dataframe is not None
        ), "You may forget to load the dataframe, use method load_dataframe first."

        if filters is None:
            return self.dataframe[features_to_select]
        query_to_filter = " and ".join(filters)
        return self.dataframe.query(query_to_filter)[features_to_select]


class BaseTransformingFeature(ABC):
    """Base class for transforming features.

    In each project, users need to create a subclass of BaseTransformingFeature. All
    transforming features in one project must be subclasses of that subclass.

    Attributes:
        name:
            (class attribute) This is the feature name that the class computes. Must be
            unique within a project.
    """

    name = ""

    def __init__(self, dependencies: List[str], raw_data: Dict):
        self.dependencies = dependencies
        self.raw_data = raw_data
        self.transformer = None

    def _transform(self, dataframe, transformer=None):
        logger.info(f"Computing feature {self.name} in pandas ...")
        # This is to make sure that _transform_ methods do not use any columns other
        # than dependencies.
        df = dataframe[self.dependencies]
        self.transformer = transformer
        if transformer is None:
            self.fit(dataframe)
        return self.transform(df)

    def fit(self, df):
        # The subclasses need to override this method if there is a need to fit the
        # transformer to data.
        self.transformer = None

    @abstractmethod
    def transform(self, df):
        raise NotImplementedError



from tabml.pipelines import BasePipeline


def run(path_to_config: str):
    pipeline = BasePipeline(path_to_config)
    pipeline.run()


def train_lgbm():
    path_to_config = "configs/lgbm_config.yaml"
    run(path_to_config)


def train_xgboost():
    path_to_config = "configs/xgboost_config.yaml"
    run(path_to_config)


def train_catboost():
    path_to_config = "configs/catboost_config.yaml"
    run(path_to_config)


if __name__ == "__main__":
    train_lgbm()



from pathlib import Path

import pytest
from qcore.asserts import AssertRaises, assert_eq

from tabml import experiment_manager
from tabml.utils.utils import write_str_to_file


class TestExperimentManager:
    @pytest.fixture(autouse=True)
    def setup_class(self, tmp_path):
        pb_str = """
            config_name: "foo"
            data_loader:
              cls_name: "bar"
              feature_config_path: "rab"
              features_to_model: ["c"]
              label_col: "d"
              train_filters: []
              validation_filters: []
            model_wrapper:
              cls_name: "bobar"
            model_analysis:
              metrics: ["a"]
              by_features: ["b"]
        """
        self.pipeline_config_path = tmp_path / "pipeline_config.pb"
        write_str_to_file(pb_str, self.pipeline_config_path)

    def test_not_create_new_run_dir(self, tmp_path):
        exp_root_dir = Path(tmp_path) / "exp"
        exp_manager = experiment_manager.ExperimentManger(
            self.pipeline_config_path, exp_root_dir=exp_root_dir
        )
        exp_manager.create_new_run_dir()
        experiment_manager.ExperimentManger(
            self.pipeline_config_path,
            should_create_new_run_dir=False,
            exp_root_dir=exp_root_dir,
        )

    def test_not_creat_run_dir_not_exist(self, tmp_path):
        exp_root_dir = Path(tmp_path) / "exp2"
        Path(exp_root_dir).mkdir(parents=True)

        with AssertRaises(IOError) as assert_raises:
            experiment_manager.ExperimentManger(
                self.pipeline_config_path,
                should_create_new_run_dir=False,
                exp_root_dir=exp_root_dir,
            )

        error_message = assert_raises.expected_exception_found
        assert_eq(
            True,
            error_message.args[0].startswith(
                "Could not find any run directory starting with"
            ),
        )



import yaml

from tabml.schemas import feature_config, pipeline_config


def parse_feature_config(yaml_path: str):
    with open(yaml_path) as f:
        config = yaml.load(f)  # config is dict
        return feature_config.FeatureConfig(**config)


def parse_pipeline_config(yaml_path: str):
    with open(yaml_path) as f:
        config = yaml.load(f)  # config is dict
        return pipeline_config.PipelineConfig(**config)



import numpy as np
import pandas as pd
from sklearn.preprocessing import LabelEncoder, StandardScaler

from tabml import data_processing, datasets
from tabml.feature_manager import BaseFeatureManager, BaseTransformingFeature

DATA_URL = (
    "https://media.githubusercontent.com/media/tiepvupsu/tabml_data/master/"
    "california_housing/housing.csv"
)


class FeatureManager(BaseFeatureManager):
    def __init__(self, feature_config_path, transformer_path=None):
        super().__init__(feature_config_path, transformer_path=transformer_path)

    def _get_base_transforming_class(self):
        return BaseHousingTransformingFeature

    def load_raw_data(self):
        full_df = datasets.load_california_housing(self.raw_data_dir)["housing"]
        full_df["house_id"] = full_df.index
        self.raw_data["full"] = full_df

    def initialize_dataframe(self):
        self.dataframe = self.raw_data["full"][
            [
                "house_id",
                "median_house_value",
                "housing_median_age",
                "total_rooms",
                "population",
                "total_bedrooms",
                "households",
                "median_income",
                "ocean_proximity",
                "latitude",
                "longitude",
            ]
        ]


class BaseHousingTransformingFeature(BaseTransformingFeature):
    pass


class FeatureIsTrain(BaseHousingTransformingFeature):
    name = "is_train"

    def transform(self, df):
        # use 90% data for training
        return df["house_id"].apply(lambda x: data_processing.hash_modulo(x, 100) < 90)


class FeatureHousingMedianAge(BaseHousingTransformingFeature):
    name = "scaled_housing_median_age"

    def fit(self, df):
        self.transformer = StandardScaler()
        train_data = df.query("is_train")[["housing_median_age"]]
        self.transformer.fit(train_data)

    def transform(self, df):
        return self.transformer.transform(df[["housing_median_age"]]).reshape(-1)


class FeatureScaledCleanTotalRooms(BaseHousingTransformingFeature):
    """Feature created by applying box plot to fix outliers then StandardScaler."""

    name = "scaled_total_rooms"

    def fit(self, df):
        self.transformer = StandardScaler()
        train_data = df.query("is_train")[["total_rooms"]]
        self.transformer.fit(train_data)

    def transform(self, df):
        return self.transformer.transform(df[["total_rooms"]]).reshape(-1)


class FeatureScaledCleanTotalBedrooms(BaseHousingTransformingFeature):
    name = "scaled_total_bedrooms"

    def fit(self, df):
        self.transformer = StandardScaler()
        train_data = df.query("is_train")[["total_bedrooms"]]
        self.transformer.fit(train_data)

    def transform(self, df):
        return self.transformer.transform(df[["total_bedrooms"]]).reshape(-1)


class FeatureScaledCleanPopulation(BaseHousingTransformingFeature):
    name = "scaled_population"

    def fit(self, df):
        self.transformer = StandardScaler()
        train_data = df.query("is_train")[["population"]]
        self.transformer.fit(train_data)

    def transform(self, df):
        return self.transformer.transform(df[["population"]]).reshape(-1)


class FeatureScaledCleanHouseholds(BaseHousingTransformingFeature):
    name = "scaled_households"

    def fit(self, df):
        self.transformer = StandardScaler()
        train_data = df.query("is_train")[["households"]]
        self.transformer.fit(train_data)

    def transform(self, df):
        return self.transformer.transform(df[["households"]]).reshape(-1)


class FeatureScaledCleanMedianIncome(BaseHousingTransformingFeature):
    name = "scaled_median_income"

    def fit(self, df):
        self.transformer = StandardScaler()
        train_data = df.query("is_train")[["median_income"]]
        self.transformer.fit(train_data)

    def transform(self, df):
        return self.transformer.transform(df[["median_income"]]).reshape(-1)


class FeatureLog10MedianHouseValue(BaseHousingTransformingFeature):
    name = "log10_median_house_value"

    def transform(self, df):
        return np.log10(df["median_house_value"])


class FeatureBucketizedLatitude(BaseHousingTransformingFeature):
    name = "bucketized_latitude"

    def transform(self, df):
        return pd.qcut(df["latitude"], q=10, labels=False)


class FeatureBucketizedLongitude(BaseHousingTransformingFeature):
    name = "bucketized_longitude"

    def transform(self, df):
        return pd.qcut(df["longitude"], q=10, labels=False)


class FeatureBucketizedLatitudeXBucketizedLongitude(BaseHousingTransformingFeature):
    name = "bucketized_latitude_X_bucketized_longitude"

    def transform(self, df):
        # TODO make feature cross more generic
        return data_processing.cross_columns(
            df, cols=["bucketized_latitude", "bucketized_longitude"]
        )


class FeatureHasdedBucketizedLatitudeXBucketizedLongitude(
    BaseHousingTransformingFeature
):
    name = "hashed_bucketized_latitude_X_bucketized_longitude"

    def transform(self, df):
        hash_bucket_size = 256
        return df["bucketized_latitude_X_bucketized_longitude"].apply(
            lambda x: data_processing.hash_modulo(x, hash_bucket_size)
        )


class FeatureEncodedOceanProximity(BaseHousingTransformingFeature):
    name = "encoded_ocean_proximity"

    def fit(self, df):
        # TODO: explicitly map raw values to coded labels
        self.transformer = LabelEncoder()
        self.transformer.fit(df["ocean_proximity"])

    def transform(self, df):
        return self.transformer.transform(df["ocean_proximity"])


def run():
    feature_config_path = "configs/feature_config.yaml"
    fm = FeatureManager(feature_config_path)
    fm.run_all()


if __name__ == "__main__":
    run()



import hashlib
from typing import Any, Dict, List, Optional, Tuple

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin


class QuantileClipper(BaseEstimator, TransformerMixin):
    def __init__(
        self,
        lower_percentile: Optional[float] = None,
        upper_percentile: Optional[float] = None,
        interpolation="linear",
    ):
        if lower_percentile is None:
            lower_percentile = 0

        if upper_percentile is None:
            upper_percentile = 1

        if lower_percentile < 0:
            raise ValueError(
                f"percentile should be in the interval [0, 1], got {lower_percentile}"
            )
        if upper_percentile > 1:
            raise ValueError(
                f"percentile should be in the interval [0, 1], got {upper_percentile}"
            )
        if lower_percentile > upper_percentile:
            raise ValueError(
                f"lower_percentile ({lower_percentile}) should not be greater than "
                f"upper_percentile ({upper_percentile})."
            )
        self.lower_percentile = lower_percentile
        self.upper_percentile = upper_percentile
        self.interpolation = interpolation
        self.lower = None
        self.upper = None

    def fit(self, X, y=None):
        self.lower = X.quantile(self.lower_percentile, interpolation=self.interpolation)
        self.upper = X.quantile(self.upper_percentile, interpolation=self.interpolation)
        return self

    def transform(self, X):
        return X.clip(lower=self.lower, upper=self.upper)


def find_boxplot_boundaries(
    col: pd.Series, whisker_coeff: float = 1.5
) -> Tuple[float, float]:
    """Findx minimum and maximum in boxplot.

    Args:
        col: a pandas serires of input.
        whisker_coeff: whisker coefficient in box plot
    """
    Q1 = col.quantile(0.25)
    Q3 = col.quantile(0.75)
    IQR = Q3 - Q1
    lower = Q1 - whisker_coeff * IQR
    upper = Q3 + whisker_coeff * IQR
    return lower, upper


class BoxplotOutlierClipper(BaseEstimator, TransformerMixin):
    def __init__(self, whisker_coeff: float = 1.5):
        self.whisker = whisker_coeff
        self.lower: float = float("-inf")
        self.upper: float = float("inf")

    def fit(self, X: pd.DataFrame, y=None):
        self.lower, self.upper = find_boxplot_boundaries(X, self.whisker)
        return self

    def transform(self, X: pd.DataFrame):
        # return array to be consistent with scikit transformer
        return np.array(X.clip(self.lower, self.upper, axis="columns"))


class CategoryEncoder:
    """A class to support encoding categorical columns.

    Attributes:
        vocab_map: a dictionary mapping each category to an integer
    """

    def __init__(self, vocab_map: Dict[Any, int]):
        self.vocab_map = vocab_map

    def get_encoded(self, series: pd.Series):
        """Maps each category to an integer based on vocabulary map.

        Unknown category will be mapped to the extra integer.
        """
        return (
            series.map(self.vocab_map)
            .fillna(max([*self.vocab_map.values()]) + 1)
            .astype(int)
        )

    @classmethod
    def from_list(cls, vocab: List[Any]):
        vocab_map = {word: i for i, word in enumerate(vocab)}
        return cls(vocab_map)

    @classmethod
    def from_lines_in_txt(cls, file_path: str):
        """Instantiates from a txt file, each line is a word."""
        with open(file_path, "r") as file:
            vocab = [line.strip() for line in file if line.strip()]
            return cls.from_list(vocab)

    @classmethod
    def from_mapping_in_csv(cls, file_path: str):
        """Instantiates from a csv file, each line is a mapping, no header."""
        df = pd.read_csv(file_path, names=["word", "code"])
        df["code"] = df["code"].astype(int)
        vocab_map = {word.strip(): code for word, code in zip(df["word"], df["code"])}
        return cls(vocab_map)


def cross_columns(df: pd.DataFrame, cols: List[str]) -> pd.Series:
    """Crosses multiple columns to create a new column.

    Resulting column has dtype = string, values of input columns are joined by "_X_"

    Args:
        df: input dataframe.
        cols: columns to cross.
    """
    return df.apply(lambda x: "_X_".join(str(x[col]) for col in cols), axis=1)


def hash_modulo(val: Any, mod: int) -> int:
    # TODO: make it faster and with more ways of hashing
    md5 = hashlib.md5()
    md5.update(str(val).encode())
    return int(md5.hexdigest(), 16) % mod



import numpy as np
import pandas as pd
from sklearn.impute import SimpleImputer
from sklearn.preprocessing import MinMaxScaler

from tabml import datasets
from tabml.feature_manager import BaseFeatureManager, BaseTransformingFeature


class FeatureManager(BaseFeatureManager):
    def __init__(self, feature_config_path, transformer_path=None):
        super().__init__(feature_config_path, transformer_path=transformer_path)

    def _get_base_transforming_class(self):
        return BaseTitanicTransformingFeature

    def load_raw_data(self):
        df_dict = datasets.load_titanic(data_dir=self.raw_data_dir)
        full_df = pd.concat(
            [df_dict["train"], df_dict["test"]], axis=0, ignore_index=True
        )
        self.raw_data["full"] = full_df

    def set_raw_data(self, raw_data_samples):
        # raw_data_samples is a list of dictionary.
        self.raw_data["full"] = pd.DataFrame(data=raw_data_samples)

    def initialize_dataframe(self):
        self.dataframe = pd.DataFrame()
        self.dataframe["passenger_id"] = self.raw_data["full"]["PassengerId"]
        self.dataframe["sibsp"] = self.raw_data["full"]["SibSp"]
        self.dataframe["parch"] = self.raw_data["full"]["Parch"]
        self.dataframe["fare"] = self.raw_data["full"]["Fare"]
        self.dataframe["age"] = self.raw_data["full"]["Age"]


class BaseTitanicTransformingFeature(BaseTransformingFeature):
    pass


class FeatureIsTrain(BaseTitanicTransformingFeature):
    name = "is_train"

    def transform(self, df):
        np.random.seed(42)
        total_train_samples = 891
        random_array = np.random.rand(len(df))
        # mask test data by 0
        random_array[total_train_samples:] = 0
        validation_size = 0.1
        return np.array(random_array > validation_size, dtype=bool)


class FeatureImputedAge(BaseTitanicTransformingFeature):
    name = "imputed_age"

    def fit(self, df):
        self.transformer = SimpleImputer(strategy="mean")
        train_data = df.query("is_train")[["age"]]
        self.transformer.fit(train_data)

    def transform(self, df):
        return self.transformer.transform(df[["age"]]).reshape(-1)


class FeatureBucketizedAge(BaseTitanicTransformingFeature):
    name = "bucketized_age"
    bins = [10, 18, 30, 40, 65]

    def transform(self, df):
        return np.digitize(df["imputed_age"], self.bins).tolist()


class FeatureSurvived(BaseTitanicTransformingFeature):
    name = "survived"

    def transform(self, df):
        return self.raw_data["full"]["Survived"]


class FeatureSex(BaseTitanicTransformingFeature):
    name = "sex"

    def transform(self, df):
        return self.raw_data["full"]["Sex"]


class FeatureCodedSex(BaseTitanicTransformingFeature):
    name = "coded_sex"

    def transform(self, df):
        return df["sex"].map({"male": 0, "female": 1})


class FeaturePclass(BaseTitanicTransformingFeature):
    name = "pclass"

    def transform(self, df):
        return self.raw_data["full"]["Pclass"]


class FeatureCodedPclass(BaseTitanicTransformingFeature):
    name = "coded_pclass"

    def transform(self, df):
        return df["pclass"].map({1: 1, 2: 2, 3: 0})


class FeatureEmbarked(BaseTitanicTransformingFeature):
    name = "embarked"

    def transform(self, df):
        embarked = self.raw_data["full"]["Embarked"]
        most_frequent = embarked.value_counts().index[0]
        return embarked.fillna(most_frequent)


class FeatureCodedEmbarked(BaseTitanicTransformingFeature):
    name = "coded_embarked"

    def transform(self, df):
        return df["embarked"].map({"C": 0, "Q": 1, "S": 2, "Unknown": 3})


class FeatureTitle(BaseTitanicTransformingFeature):
    name = "title"

    def transform(self, df):
        title = self.raw_data["full"]["Name"].str.extract(
            r" ([A-Za-z]+)\.", expand=False
        )
        title = title.replace(
            [
                "Dr",
                "Rev",
                "Col",
                "Major",
                "Countess",
                "Sir",
                "Jonkheer",
                "Lady",
                "Capt",
                "Don",
                "Dona",
            ],
            "Others",
        )
        title = title.replace("Ms", "Miss")
        title = title.replace("Mme", "Mrs")
        title = title.replace("Mlle", "Miss")
        return title


class FeatureCodedTitle(BaseTitanicTransformingFeature):
    name = "coded_title"

    def transform(self, df):
        return df["title"].map(
            {"Master": 0, "Miss": 1, "Mr": 2, "Mrs": 3, "Others": 4, "Unknown": 5}
        )


class FeatureMinMaxScaledAge(BaseTitanicTransformingFeature):
    name = "min_max_scaled_age"

    def fit(self, df):
        self.transformer = MinMaxScaler()
        train_data = df.query("is_train")[["imputed_age"]]
        self.transformer.fit(train_data)

    def transform(self, df):
        return self.transformer.transform(df[["imputed_age"]]).reshape(-1)


def run(transformer_path=None):
    feature_config_path = "configs/feature_config.yaml"
    fm = FeatureManager(feature_config_path, transformer_path)
    fm.run_all()


if __name__ == "__main__":
    run()



import typing

import numpy as np
from qcore.asserts import AssertRaises, assert_eq

from tabml import metrics


class DummyMetric1(metrics.BaseMetric):
    name = "dummy1"
    score_names = ["a"]

    def _compute_scores(
        self, labels: typing.Collection, preds: typing.Collection
    ) -> typing.List[float]:
        return [0]


class DummyMetric2(metrics.BaseMetric):
    name = "dummy1"
    is_higher_better = True
    score_names = ["a"]

    def _compute_scores(
        self, labels: typing.Collection, preds: typing.Collection
    ) -> typing.List[float]:
        return [0]


class DummyMetric3(metrics.BaseMetric):
    name = "dummy1"
    is_higher_better = True
    score_names = ["a"]

    def _compute_scores(
        self, labels: typing.Collection, preds: typing.Collection
    ) -> typing.List[float]:
        return [0, 0]


class TestBaseMetric:
    def test_raise_ve_if_is_higher_better_not_defind(self):
        with AssertRaises(ValueError) as assert_raises:
            DummyMetric1().compute_scores([0], [0])

        error_message = assert_raises.expected_exception_found
        assert_eq(
            True,
            error_message.args[0].startswith(
                "Subclasses must define is_higher_better."
            ),
        )

    def test_raise_ve_if_labels_and_preds_have_diff_lens(self):
        with AssertRaises(ValueError) as assert_raises:
            DummyMetric2().compute_scores([0, 0], [0])

        error_message = assert_raises.expected_exception_found
        assert_eq(
            True,
            error_message.args[0].startswith("labels (len = 2) and preds (len = 1)"),
        )

    def test_score_names_and_compute_scores_have_diff_lens(self):
        with AssertRaises(ValueError) as assert_raises:
            DummyMetric3().compute_scores([0], [0])

        error_message = assert_raises.expected_exception_found
        assert_eq(
            True,
            error_message.args[0].startswith(
                "self.score_names (['a']) and scores ([0, 0]) must have the same length"
            ),
        )


class TestComputeMaxF1AndThreshold:
    def test_raise_if_labels_not_binary(self):
        labels = [0, 2]
        preds = [0, 1]
        with AssertRaises(ValueError) as assert_raises:
            metrics.compute_maxf1_stats(labels, preds)

        error_message = assert_raises.expected_exception_found
        assert_eq(
            True,
            error_message.args[0].startswith(
                "Input must contain only binary values, got "
            ),
        )

    def test_1(self):
        labels = [0, 0, 1, 1]
        preds = [0.1, 0.6991, 0.7009, 0.9]
        expected_max_f1 = 1
        expected_max_f1_threshold = 0.7
        expected_max_f1_precision = 1
        expected_max_f1_recall = 1
        max_f1, threshold, precision, recall = metrics.compute_maxf1_stats(
            labels, preds
        )
        assert_eq(expected_max_f1, max_f1)
        assert_eq(expected_max_f1_threshold, threshold, tolerance=1e-6)
        assert_eq(expected_max_f1_precision, precision)
        assert_eq(expected_max_f1_recall, recall)

    def test_2(self):
        labels = [0, 0, 1, 1]
        preds = [0.9, 0.7, 0.3, 0.1]
        expected_max_f1 = 2 / 3  # pr = 1/2, recall = 1
        expected_max_f1_precision = 1 / 2
        expected_max_f1_recall = 1
        max_f1, _, precision, recall = metrics.compute_maxf1_stats(labels, preds)
        assert_eq(expected_max_f1, max_f1)
        assert_eq(expected_max_f1_precision, precision)
        assert_eq(expected_max_f1_recall, recall)


class TestRocAuc:
    def test_return_nan_if_uniqe_label(self):
        got = metrics.RocAreaUnderTheCurve().compute_scores([0, 0], [0.5, 0.5])
        expected = {"roc_auc": np.nan}
        assert_eq(expected, got)


class TestF1Score:
    def test_computation(self):
        labels = [1, 0, 1, 1, 0]
        preds = [0, 0, 1, 1, 1]
        expected = {"f1": 2 / 3, "precision": 2 / 3, "recall": 2 / 3}
        got = metrics.F1Score().compute_scores(labels, preds)
        assert_eq(expected, got)


class TestSmape:
    def test_computation(self):
        labels = [1, 100, 1000]
        preds = [10, 10, 800]
        expected = 200 / 3 * (9 / 11 + 90 / 110 + 200 / 1800)
        got = metrics.SMAPE().compute_scores(labels, preds)["smape"]
        np.testing.assert_almost_equal(expected, got)


def test_metrics_have_unique_name():
    metric_by_name = metrics.get_instantiated_metric_dict()
    metric_names = metric_by_name.keys()
    assert_eq(len(metric_names), len(set(metric_names)))



from abc import ABC, abstractmethod
from typing import Dict, Iterable, Tuple

import mlflow
import numpy as np
import shap
from catboost import CatBoostClassifier, CatBoostRegressor
from lightgbm import LGBMClassifier, LGBMRegressor
from xgboost import XGBClassifier, XGBRegressor

from tabml.data_loaders import BaseDataLoader
from tabml.schemas import pipeline_config
from tabml.utils import factory, utils
from tabml.utils.logger import boosting_logger_eval
from tabml.utils.utils import save_as_pickle

MLFLOW_AUTOLOG = {
    "sklearn": mlflow.sklearn.autolog(),
    "lightgbm": mlflow.lightgbm.autolog(),
    "xgboost": mlflow.xgboost.autolog(),
    "catboost": None,
}


class BaseModelWrapper(ABC):
    mlflow_model_type = ""

    def __init__(self, params=pipeline_config.ModelWrapper()):
        self.model = None
        # Parameters for model instantiating
        self.model_params = params.model_params
        # Parameters for model training
        self.fit_params = params.fit_params

    def fit(self, data_loader: BaseDataLoader, model_dir: str):
        pass

    @abstractmethod
    def predict(self, data) -> Iterable:
        """Predicts data inputs."""
        raise NotImplementedError

    def predict_proba(self, data) -> Iterable:
        """Predicts probability of data inputs.

        Only applied to binary classification problems, the second value in
        probability (prob of positive) is chosen.
        """
        raise NotImplementedError

    @abstractmethod
    def load_model(self, model_path: str):
        raise NotImplementedError

    def get_feature_importance(self, input_data) -> Dict[str, float]:
        """Computes feature importance for each feature based on an input data.

        Most of models are supported by SHAP (https://github.com/slundberg/shap). For
        unsupported models, please override this method by a workable solution.
        """
        explainer = shap.Explainer(self.model)
        shap_values = explainer(input_data)

        def _get_shap_values_one_sample(shap_values, index: int):
            # For LightGBM and XGBoost, shap_values[index].values is a 2d array
            # representing logit of two classes. They are basically negative of each
            # other, we only need one.
            # Related issue https://github.com/slundberg/shap/issues/526.
            if len(shap_values[index].values.shape) == 2:  # binary classification
                return shap_values[index].values[:, 0]
            assert len(shap_values[index].values.shape) == 1, len(
                shap_values[index].values
            )
            return shap_values[index].values

        feature_importances = np.mean(
            [
                np.abs(_get_shap_values_one_sample(shap_values, i))
                for i in range(len(shap_values))
            ],
            axis=0,
        ).tolist()

        feature_names = input_data.columns.tolist()
        feature_importance_dict = dict(zip(feature_names, feature_importances))
        return feature_importance_dict


class BaseSklearnModelWrapper(BaseModelWrapper):
    """A common model wrapper for scklearn-like models."""

    mlflow_model_type = "sklearn"

    def __init__(self, params=pipeline_config.ModelWrapper()):
        super().__init__(params)
        self.save_model_name = "model_0"
        self.model = factory.create(params.model_cls)(**self.model_params)

    def fit(self, data_loader: BaseDataLoader, model_dir: str):
        assert (
            data_loader.label_col is not None
        ), "data_loader.label_col must be declared in BaseDataLoader subclasses."
        train_feature, train_label = data_loader.get_train_data_and_label()

        self.model.fit(X=train_feature, y=train_label, **self.fit_params)
        save_as_pickle(self.model, model_dir, self.save_model_name)

    def load_model(self, model_path: str):
        self.model = utils.load_pickle(model_path)

    def predict(self, data):
        return self.model.predict(data)

    def predict_proba(self, data) -> Iterable:
        return self.model.predict_proba(data)[:, 1]


class BaseBoostingModelWrapper(BaseModelWrapper):
    """A common model wrapper for boosting models.

    Boosting models: LightGBM, XGBoost, CatBoost."""

    def __init__(self, params=pipeline_config.ModelWrapper()):
        super().__init__(params)
        self.save_model_name = "model_0"

    @abstractmethod
    def _get_fit_params(self, train_data: Tuple, val_data: Tuple) -> Dict:
        raise NotImplementedError

    def fit(self, data_loader: BaseDataLoader, model_dir: str):
        assert (
            data_loader.label_col is not None
        ), "data_loader.label_col must be declared in BaseDataLoader subclasses."
        train_feature, train_label = data_loader.get_train_data_and_label()
        val_data = data_loader.get_val_data_and_label()

        fit_params = self._get_fit_params((train_feature, train_label), val_data)

        # breakpoint()
        self.model.fit(X=train_feature, y=train_label, **fit_params)
        save_as_pickle(self.model, model_dir, self.save_model_name)


class BaseLgbmModelWrapper(BaseBoostingModelWrapper):
    mlflow_model_type = "lightgbm"

    def __init__(self, params=pipeline_config.ModelWrapper()):
        super().__init__(params)
        self.model = self.build_model()

    @abstractmethod
    def build_model(self):
        raise NotImplementedError

    def predict(self, data):
        return self.model.predict(data)

    def load_model(self, model_path: str):
        self.model = utils.load_pickle(model_path)

    def _get_fit_params(self, train_data: Tuple, val_data: Tuple) -> Dict:
        fit_params = {
            "eval_set": [train_data, val_data],
            "eval_names": ["train", "val"],
            "callbacks": [boosting_logger_eval(model="lgbm")],
            **self.fit_params,
        }
        return fit_params


class LgbmClassifierModelWrapper(BaseLgbmModelWrapper):
    def build_model(self):
        return LGBMClassifier(**self.model_params)

    def predict_proba(self, data) -> Iterable:
        return self.model.predict_proba(data)[:, 1]


class LgbmRegressorModelWrapper(BaseLgbmModelWrapper):
    def build_model(self):
        return LGBMRegressor(**self.model_params)


class BaseXGBoostModelWrapper(BaseBoostingModelWrapper):
    mlflow_model_type = "xgboost"

    def __init__(self, params=pipeline_config.ModelWrapper):
        super().__init__(params)
        self.tree_method = "gpu_hist" if utils.is_gpu_available() else "auto"
        self.model = self.build_model()

    @abstractmethod
    def build_model(self):
        pass

    def predict(self, data):
        return self.model.predict(data)

    def load_model(self, model_path: str):
        self.model = utils.load_pickle(model_path)

    def _get_fit_params(self, train_data, val_data):
        fit_params = {
            "eval_set": [train_data, val_data],
            "callbacks": [boosting_logger_eval(model="xgboost")],
            **self.fit_params,
        }
        return fit_params


class XGBoostRegressorModelWrapper(BaseXGBoostModelWrapper):
    def build_model(self):
        return XGBRegressor(tree_method=self.tree_method, **self.model_params)


class XGBoostClassifierModelWrapper(BaseXGBoostModelWrapper):
    def build_model(self):
        return XGBClassifier(tree_method=self.tree_method, **self.model_params)

    def predict_proba(self, data) -> Iterable:
        return self.model.predict_proba(data)[:, 1]


class BaseCatBoostModelWrapper(BaseBoostingModelWrapper):
    mlflow_model_type = "catboost"

    def __init__(self, params=pipeline_config.ModelWrapper()):
        super().__init__(params)
        self.task_type = "GPU" if utils.is_gpu_available() else "CPU"
        self.model = self.build_model()

    @abstractmethod
    def build_model(self):
        pass

    def predict(self, data):
        return self.model.predict(data)

    def load_model(self, model_path: str):
        self.model = utils.load_pickle(model_path)

    def _get_fit_params(self, train_data, val_data):
        fit_params = {"eval_set": [val_data]}

        return fit_params


class CatBoostClassifierModelWrapper(BaseCatBoostModelWrapper):
    def build_model(self):
        return CatBoostClassifier(**self.model_params)

    def predict_proba(self, data) -> Iterable:
        return self.model.predict_proba(data)[:, 1]


class CatBoostRegressorModelWrapper(BaseCatBoostModelWrapper):
    def build_model(self):
        return CatBoostRegressor(task_type=self.task_type, **self.model_params)


def write_model_wrapper_subclasses_to_file(
    base_cls=BaseModelWrapper, md_path="model_wrapper_map.md"
):
    lines = ["# Inheritance map\n", "\n"]
    level = 0
    stack = [(base_cls, level)]
    while stack:
        node, level = stack.pop()
        lines.append("    " * level + f"- {node}\n")
        stack.extend((child, level + 1) for child in node.__subclasses__())

    with open(md_path, "w") as f:
        f.writelines(lines)


if __name__ == "__main__":
    # Show the subclasses of BaseModelWrapper
    write_model_wrapper_subclasses_to_file()



from distutils.core import setup

import setuptools

setup(
    name="tabml",
    packages=setuptools.find_packages(),
    version="0.1.13",
    license="apache-2.0",
    description="A package for machine learning with tabular data",
    author="Tiep Vu",
    author_email="vuhuutiep@gmail.com",
    url="https://github.com/tiepvupsu/tabml",
    download_url="https://github.com/tiepvupsu/tabml/archive/refs/tags/v0_1_13.tar.gz",
    keywords=["Machine Learning", "Tabular"],
    install_requires=[
        "catboost>=0.26.1",
        "GPUtil>=1.4.0",
        "kaggle>=1.5.6",
        "lightgbm>=2.3.1",
        "loguru>=0.5.1",
        "mlflow>=1.20.1",
        "numpy>=1.18.5",
        "pandas>=1.3.1",
        "pandas-profiling>=2.9.0",
        "pydantic>=1.8.2",
        "pyyaml>=5.4.1",
        "scikit-learn>=0.24.2",
        "scipy>=1.6.2",
        "shap>=0.39.0",
        "termgraph>=0.4.2",
        "xgboost>=1.4.2",
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.8",
    ],
)



from typing import Dict, List, Union

import pydantic


class DataLoader(pydantic.BaseModel):
    cls_name: str = "tabml.data_loaders.BaseDataLoader"  # name of DataLoader class

    # path to feature_config
    feature_config_path: str = ""

    # list of features going in to the model
    features_to_model: List[str] = []

    # filters to determine training, validation, and submission dataset. For each
    # dataset, rows that are met all conditions will be selected.
    train_filters: List[str] = []
    validation_filters: List[str] = []
    submission_filters: List[str] = []

    # name of label column
    label_col: str = ""


class ModelWrapper(pydantic.BaseModel):
    cls_name: str = "tabml.data_loaders.BaseDataLoader"  # name of model_wrapper class
    model_cls: str = ""  # a siklearn-like model class
    model_params: Dict = {}
    fit_params: Dict = {}


class ModelAnalysis(pydantic.BaseModel):
    # list of metrics to compute.
    # Each element must be the name of a subclass of tabml.metrics.BaseMetric
    metrics: List[str] = []

    # list of features for analysis
    by_features: List[str] = []

    # This field specifies the label for analysis. In some cases, we use a
    # transformation of the true label for training but need to evaluate on the
    # original label.
    by_label: Union[str, None] = None

    # To avoid expensive computations, e.g. computing SHAP values or inference,
    # a subset of training data could be used instead. This field specifies size
    # of that subset.
    training_size: Union[int, None] = None


class PipelineConfig(pydantic.BaseModel):
    config_name: str
    data_loader: DataLoader
    model_wrapper: ModelWrapper
    model_analysis: ModelAnalysis



import sys

from loguru import logger

logger.add(
    sys.stderr, format="{time} {level} {message}", filter="my_module", level="INFO"
)


def lgbm_format_eval_result(value, show_stdv=True):
    if len(value) == 4:
        return "{0}'s {1}:{2:.5f}".format(value[0], value[1], value[2])
    elif len(value) == 5:
        if show_stdv:
            return "{0}'s {1}:{2:.5f}+{3:.5f}".format(
                value[0], value[1], value[2], value[4]
            )
        return "{0}'s {1}:{2:.5f}".format(value[0], value[1], value[2])
    raise ValueError("Wrong metric value")


def xgboost_format_eval_result(value, show_stdv=True):
    if len(value) == 2:
        return "{0}:{1:.5f}".format(value[0], value[1])
    if len(value) == 3:
        if show_stdv:
            return "{0}:{1:.5f}+{2:.5f}".format(value[0], value[1], value[2])
        return "{0}:{1:.5f}".format(value[0], value[1])
    raise ValueError("wrong metric value")


def boosting_logger_eval(model: str, period: int = 1, show_stdv: bool = True):
    """Creates a callback that prints the evaluation results for lgbm.

    Args:
        model (str): model name, lgbm or xgb
        period (int, optional):
            The period to print the evaluation results.
            Defaults to 1
        show_stdv (bool, optional):
            Whether to show stdv (if provided).
            Default to True

    Returns:
        function: The callback that prints the evaluation results
            every ``period`` iteration(s).
    """
    if model == "lgbm":
        format_result_func = lgbm_format_eval_result
    elif model == "xgboost":
        format_result_func = xgboost_format_eval_result
    else:
        raise TypeError(f"Unexpected model {model}. Only accept lgbm or xgb model.")

    def _callback(env):
        if (
            period > 0
            and env.evaluation_result_list
            and (env.iteration + 1) % period == 0
        ):
            result = "\t".join(
                [format_result_func(x, show_stdv) for x in env.evaluation_result_list]
            )
            logger.info(f"[{env.iteration}] {result}")

    return _callback



import os
import pickle
import random
import string
import subprocess
import tempfile
import typing
from collections import Counter
from pathlib import Path
from typing import Any, Collection

import GPUtil

from tabml.utils.logger import logger


def random_string(length: int = 10) -> str:
    """Returns a random string of lowercase letters and digits with a given length.

    Args:
        length: output length

    Returns:
        a random string with length being `length`
    """
    return "".join(
        [random.choice(string.ascii_letters + string.digits) for n in range(length)]
    )


def write_str_to_file(a_str: str, filename: str) -> None:
    """Writes a string to a file."""
    with open(filename, "w") as str_file:
        str_file.write(a_str)


def check_uniqueness(items: Collection) -> None:
    """Checks if an array containing unique elements.

    Args:
        items: A list of objects.

    Returns:
        Does not return anything. If this function passes, it means that all objects
        are unique.

    Raises:
        Assertion error with list of duplicate objects.
    """
    counter = Counter(items)
    duplicates = {item: count for item, count in counter.items() if count > 1}
    assert not duplicates, f"There are duplicate objects in the list: {duplicates}."


def show_feature_importance(data: typing.Dict[str, float]) -> None:
    """
    Shows feature importance in terminal.

    Importances are shown in desecending order.
    Note that termgraph (a tool for visualize graph in terminal) only shows meaningful
    graphs with positive values. Fortunately, XGB model only outputs positive feature
    importances. If LGBM and Keras, if any, have negative feature importance, then a
    quick modification is to visualize absolute values. In fact, the magnitude of
    feature importances are more _important_ than their actual values.

    Args:
        data: a dictionary of {feature: imporatnce}

    Raises:
        ValueError if any importance is negative.
    """
    assert data is not None, "input dictionary can not be empty"
    tmp_file = tempfile.NamedTemporaryFile()
    # sort feature by desecending importance
    feature_importance_tuples = sorted(data.items(), key=lambda x: -x[1])

    if feature_importance_tuples[-1][-1] < 0:
        raise ValueError(
            f"All feature importances need to be non-negative, got data = {data}"
        )

    # write to file
    # TODO: find a way to visualize data directly rather than saving it in an
    # intermediate file.
    with open(tmp_file.name, "w") as fout:
        for feature, importance in feature_importance_tuples:
            fout.write(f"{feature}, {importance}\n")

    logger.info("Feature importance:")
    logger.info(subprocess.getoutput(f"termgraph {tmp_file.name}"))


def save_as_pickle(an_object: Any, path: str, filename: str) -> None:
    """Saves an object as a pickle file.

    Args:
        an_object: A python object. Can be list, dict etc.
        path: The path where to save.
        filename: The filename
    """
    if not Path(path).exists():
        Path(path).mkdir(parents=True, exist_ok=True)

    file_path = Path(path) / filename
    with open(file_path, "wb") as handle:
        pickle.dump(an_object, handle)
    logger.info(f"File is saved successfully to {file_path}.")


def load_pickle(path_to_obj):
    # open a file, where you stored the pickled data
    file = open(path_to_obj, "rb")
    # dump information to that file
    obj = pickle.load(file)
    # close the file
    file.close()

    return obj


def is_gpu_available():
    return (
        len(
            GPUtil.getAvailable(
                order="first",
                limit=1,
                maxLoad=0.5,
                maxMemory=0.5,
                includeNan=False,
                excludeID=[],
                excludeUUID=[],
            )
        )
        > 0
    )


def change_working_dir_pytest(func):
    # https://stackoverflow.com/a/62055409/11871829
    def apply(request):
        os.chdir(request.fspath.dirname)
        func()
        os.chdir(request.config.invocation_dir)

    return apply



from pathlib import Path

import numpy as np
import pandas as pd
import pytest

from tabml.config_helpers import parse_pipeline_config
from tabml.data_loaders import BaseDataLoader
from tabml.feature_manager import BaseFeatureManager
from tabml.utils.utils import write_str_to_file


class TestBaseDataLoader:
    @pytest.fixture(autouse=True)
    def setup_class(self, tmp_path):
        # feature manager config
        dataset_dir = tmp_path / "dataset"
        fm_pb_str = f"""
        raw_data_dir: "{dataset_dir}"
        dataset_name: "dummy"
        base_features:
          - name: "a"
            dtype: INT32
        transforming_features:
          - name: "b"
            index: 1
            dtype: INT32
          - name: "label"
            index: 2
            dtype: INT32
          - name: "is_train"
            index: 3
            dtype: BOOL
          - name: "is_validation"
            index: 4
            dtype: BOOL
        """
        fm_pb_path = tmp_path / "feature_config.yaml"
        write_str_to_file(fm_pb_str, fm_pb_path)

        # create fake data
        df = pd.DataFrame(
            data={
                "a": [1, 2, 3, 4, 5],
                "b": [6, 7, 8, 9, 0],
                "c": [-1, -1, -1, -1, -1],
                "label": [0, 1, 1, 0, 1],
                "is_train": [True, False, True, True, False],
                "is_validation": [False, True, False, False, True],
            }
        )
        dataset_path = BaseFeatureManager(fm_pb_path).get_dataset_path()
        Path(dataset_path).parent.mkdir(parents=True)
        df.to_csv(dataset_path, index=False)

        # pipeline config
        pipeline_config = f"""
        config_name: "dummy"
        data_loader:
            cls_name: "tabml.data_loaders.BaseDataLoader"
            feature_config_path: "{fm_pb_path}"
            label_col: "label"
            features_to_model: ["a", "b"]
            train_filters: ["is_train"]
            validation_filters: ["is_validation"]
        model_wrapper:
            cls_name: "a"
        model_analysis:
            metrics: ["foo"]
            by_features: ["bar"]
            by_label: "bar"
            training_size: 50
        """

        pipeline_config_path = tmp_path / "pipeline_config.yaml"
        write_str_to_file(pipeline_config, pipeline_config_path)
        self.config = parse_pipeline_config(pipeline_config_path)

    def test_get_train_data_and_label(self):
        data_loader = BaseDataLoader(self.config.data_loader)
        got_data, got_label = data_loader.get_train_data_and_label()
        expected_data = pd.DataFrame(data={"a": [1, 3, 4], "b": [6, 8, 9]})
        expected_label = pd.Series(data={"label": [0, 1, 0]})
        # since index of expected_ and got_ are different, it's better to compare
        # their values.
        np.array_equal(expected_data.values, got_data.values)
        np.array_equal(expected_label.values, got_label.values)

    def test_get_validation_data_and_label(self):
        data_loader = BaseDataLoader(self.config.data_loader)
        got_data, got_label = data_loader.get_val_data_and_label()
        expected_data = pd.DataFrame(data={"a": [2, 5], "b": [7, 0]})
        expected_label = pd.Series(data={"label": [1, 1]})
        # since index of expected_ and got_ are different, it's better to compare
        # their values.
        np.array_equal(expected_data.values, got_data.values)
        np.array_equal(expected_label.values, got_label.values)
